import abc
import binascii
import datetime
import enum
import io
import re
from collections.abc import Iterator
from typing import Literal, overload

TimestampType = datetime.datetime | datetime.timedelta | None
TimezoneType = datetime.tzinfo | Literal['local'] | None

TZ_LOCAL: Literal['local'] = 'local'

class TimestampFormat(enum.Enum):
    """Describes a type of timestamp. ABSOLUTE is referring to UNIX time
    (seconds since epoch). RELATIVE is seconds since start of log, or time
    since last frame depending of the contents of the log file. MISSING means
    that no timestamps is present in the log."""
    ABSOLUTE = 1
    RELATIVE = 2
    MISSING = 3


class DataFrame:
    """Container for a parsed log entry (ie. a CAN frame)."""

    def __init__(self, channel: str,
                 frame_id: int,
                 is_extended_frame: bool,
                 data: bytes,
                 is_remote_frame: bool,
                 timestamp: TimestampType,
                 timestamp_format: TimestampFormat):
        """Constructor for DataFrame

        :param channel: A string representation of the channel, eg. 'can0'
        :param frame_id: The numeric CAN frame ID :param data: The actual data
        :param timestamp: A timestamp, datetime.datetime if absolute, or
            datetime.timedelta if relative, None if missing
        :param timestamp_format: The format of the timestamp
        : """
        self.channel = channel
        self.frame_id = frame_id
        self.is_extended_frame = is_extended_frame
        self.data = bytes(data)
        self.is_remote_frame = is_remote_frame
        self.timestamp = timestamp
        self.timestamp_format = timestamp_format

    def __repr__(self) -> str:
        attrs = ', '.join(f'{a} = {getattr(self, a)!r}' for a in self.__dict__.keys())
        return f'{type(self).__name__}({attrs})'


class BasePattern:

    pattern: re.Pattern[str]

    def match(self, line: str) -> DataFrame | None:
        mo = self.pattern.match(line)
        if mo:
            return self.unpack(mo)

        return None

    @abc.abstractmethod
    def unpack(self, match_object: re.Match[str]) -> DataFrame | None:
        raise NotImplementedError()


class CandumpBasePattern(BasePattern):

    def unpack(self, match_object: re.Match[str]) -> DataFrame:
        channel = match_object.group('channel')
        frame_id = int(match_object.group('can_id'), 16)
        is_extended_frame = len(match_object.group('can_id')) > 3
        data = match_object.group('can_data')
        if data in {'remote request', 'R'}:
            is_remote_frame = True
            data = bytes(0)
        else:
            is_remote_frame = False
            data = data.replace(' ', '')
            data = binascii.unhexlify(data)
        timestamp, timestamp_format = self.parse_timestamp(match_object)

        return DataFrame(channel=channel, frame_id=frame_id, is_extended_frame=is_extended_frame, data=data, is_remote_frame=is_remote_frame, timestamp=timestamp, timestamp_format=timestamp_format)

    @abc.abstractmethod
    def parse_timestamp(self, match_object: re.Match[str]) -> tuple[TimestampType, TimestampFormat]:
        raise NotImplementedError()


class CandumpDefaultPattern(CandumpBasePattern):
    #candump vcan0
    # vcan0  1F0   [8]  00 00 00 00 00 00 1B C1
    #candump vcan0 -a
    # vcan0  1F0   [8]  00 00 00 00 00 00 1B C1   '.......Á'
    #(Ignore anything after the end of the data to work with candump's ASCII decoding)
    pattern = re.compile(
        r'^\s*?(?P<channel>[a-zA-Z0-9]+)\s+(?P<can_id>[0-9A-F]+)\s+\[\d+\]\s*(?P<can_data>remote request|[0-9A-F ]*).*?$')

    def parse_timestamp(self, match_object: re.Match[str]) -> tuple[TimestampType, TimestampFormat]:
        timestamp = None
        timestamp_format = TimestampFormat.MISSING
        return timestamp, timestamp_format


class CandumpTimestampedPattern(CandumpBasePattern):
    #candump vcan0 -tz
    # (000.000000)  vcan0  0C8   [8]  F0 00 00 00 00 00 00 00
    #candump vcan0 -tz -a
    # (000.000000)  vcan0  0C8   [8]  31 30 30 2E 35 20 46 4D   '100.5 FM'
    #(Ignore anything after the end of the data to work with candump's ASCII decoding)
    pattern = re.compile(
        r'^\s*?\((?P<timestamp>[\d.]+)\)\s+(?P<channel>[a-zA-Z0-9]+)\s+(?P<can_id>[0-9A-F]+)\s+\[\d+\]\s*(?P<can_data>remote request|[0-9A-F ]*).*?$')

    def __init__(self, tz: TimezoneType) -> None:
        if tz == TZ_LOCAL:
            tz = None
            self.convert_to_timezone_aware_object = True
        else:
            self.convert_to_timezone_aware_object = False
        self.tz = tz

    def parse_timestamp(self, match_object: re.Match[str]) -> tuple[TimestampType, TimestampFormat]:
        seconds = float(match_object.group('timestamp'))
        timestamp: datetime.timedelta | datetime.datetime
        if seconds < 662688000:  # 1991-01-01 00:00:00, "Released in 1991, the Mercedes-Benz W140 was the first production vehicle to feature a CAN-based multiplex wiring system."
            timestamp = datetime.timedelta(seconds=seconds)
            timestamp_format = TimestampFormat.RELATIVE
        else:
            timestamp = datetime.datetime.fromtimestamp(seconds, self.tz)
            if self.convert_to_timezone_aware_object:
                timestamp = timestamp.astimezone()
            timestamp_format = TimestampFormat.ABSOLUTE

        return timestamp, timestamp_format


class CandumpDefaultLogPattern(CandumpBasePattern):
    # (1579857014.345944) can2 486#82967A6B006B07F8
    # (1613656104.501098) can2 14C##16A0FFE00606E022400000000000000A0FFFF00FFFF25000600000000000000FE
    pattern = re.compile(
        r'^\s*?\((?P<timestamp>[\d.]+?)\)\s+?(?P<channel>[a-zA-Z0-9]+)\s+?(?P<can_id>[0-9A-F]+?)#(#[0-9A-F])?(?P<can_data>R|([0-9A-Fa-f]{2})*)(\s+[RT])?$')

    def __init__(self, tz: TimezoneType) -> None:
        if tz == TZ_LOCAL:
            tz = None
            self.convert_to_timezone_aware_object = True
        else:
            self.convert_to_timezone_aware_object = False
        self.tz = tz

    def parse_timestamp(self, match_object: re.Match[str]) -> tuple[TimestampType, TimestampFormat]:
        timestamp = datetime.datetime.fromtimestamp(float(match_object.group('timestamp')), self.tz)
        if self.convert_to_timezone_aware_object:
            timestamp = timestamp.astimezone()
        timestamp_format = TimestampFormat.ABSOLUTE
        return timestamp, timestamp_format


class CandumpAbsoluteLogPattern(CandumpBasePattern):
    #candump vcan0 -tA
    # (2020-12-19 12:04:45.485261)  vcan0  0C8   [8]  F0 00 00 00 00 00 00 00
    #candump vcan0 -tA -a
    # (2020-12-19 12:04:45.485261)  vcan0  0C8   [8]  31 30 30 2E 35 20 46 4D   '100.5 FM'
    #(Ignore anything after the end of the data to work with candump's ASCII decoding)
    pattern = re.compile(
        r'^\s*?\((?P<timestamp>\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d+)\)\s+(?P<channel>[a-zA-Z0-9]+)\s+(?P<can_id>[0-9A-F]+)\s+\[\d+\]\s*(?P<can_data>remote request|[0-9A-F ]*).*?$')

    def parse_timestamp(self, match_object: re.Match[str]) -> tuple[TimestampType, TimestampFormat]:
        timestamp = datetime.datetime.strptime(match_object.group('timestamp'), "%Y-%m-%d %H:%M:%S.%f")
        timestamp_format = TimestampFormat.ABSOLUTE
        return timestamp, timestamp_format


class PCANTracePatternV10(BasePattern):
    """
    Reference for PCAN trace patterns: https://www.peak-system.com/produktcd/Pdf/English/PEAK_CAN_TRC_File_Format.pdf
    1) 1841 0001 8 00 00 00 00 00 00 00 00

    >>> PCANTracePatternV10().match(" 1) 1841 0001 8 00 00 00 00 00 00 00 00") #doctest: +ELLIPSIS
    <logreader.DataFrame object at ...>
    """
    pattern = re.compile(
        r'^\s*?\d+\)\s*?(?P<timestamp>\d+)\s+(?P<can_id>[0-9A-F]+)\s+(?P<dlc>[0-9])\s+(?P<can_data>RTR|[0-9A-F ]*)$')

    def unpack(self, match_object: re.Match[str]) -> DataFrame | None:
        channel = self.parse_channel(match_object)
        frame_id = int(match_object.group('can_id'), 16)
        is_extended_frame = len(match_object.group('can_id')) > 4
        data, is_remote_frame = self.parse_data(match_object)
        millis = float(match_object.group('timestamp'))
        # timestamp = datetime.datetime.strptime(match_object.group('timestamp'), "%Y-%m-%d %H:%M:%S.%f")
        timestamp = datetime.timedelta(milliseconds=millis)
        timestamp_format = TimestampFormat.RELATIVE

        return DataFrame(channel=channel, frame_id=frame_id, is_extended_frame=is_extended_frame, data=data, is_remote_frame=is_remote_frame, timestamp=timestamp, timestamp_format=timestamp_format)

    def parse_channel(self, match_object: re.Match[str]) -> str:
        return 'pcanx'

    def parse_data(self, match_object: re.Match[str]) -> tuple[bytes, bool]:
        data = match_object.group('can_data')
        if data == 'RTR':
            is_remote_frame = True
            data = bytes(0)
        else:
            is_remote_frame = False
            data = bytes.fromhex(data)

        return data, is_remote_frame


class PCANTracePatternV11(PCANTracePatternV10):
    """
    Adds "Type" 'Rx' column to 1.0 and 1/10 microsecond resolution
    1)      6357.2 Rx        0401  8    00 00 00 00 00 00 00 00

    >>> PCANTracePatternV11().match("  1)      6357.2  Rx        0401  8    00 00 00 00 00 00 00 00") #doctest: +ELLIPSIS
    <logreader.DataFrame object at ...>
    """
    pattern = re.compile(
        r'^\s*?\d+\)\s*?(?P<timestamp>\d+.\d+)\s+(?P<type>\w+)\s+(?P<can_id>[0-9A-F]+)\s+(?P<dlc>[0-9])\s+(?P<can_data>RTR|[0-9A-F ]*)$')

    def unpack(self, match_object: re.Match[str]) -> DataFrame | None:
        if match_object.group('type') in ('Error', 'Warng'):  # yes, they really spell Warning without the 'in'
            return None

        return super().unpack(match_object)

class PCANTracePatternV12(PCANTracePatternV11):
    """
    Adds "Bus" column and 1 microsecond resolution to 1.1
    1)      6357.213 1  Rx        0401  8    00 00 00 00 00 00 00 00

    >>> PCANTracePatternV12().match("  1)      6357.213 1  Rx        0401  8    00 00 00 00 00 00 00 00") #doctest: +ELLIPSIS
    <logreader.DataFrame object at ...>
    """
    pattern = re.compile(
        r'^\s*?\d+\)\s*?(?P<timestamp>\d+.\d+)\s+(?P<channel>[0-9])\s+(?P<type>\w+)\s+(?P<can_id>[0-9A-F]+)\s+(?P<dlc>[0-9])\s+(?P<can_data>RTR|[0-9A-F ]*)$')

    def parse_channel(self, match_object: re.Match[str]) -> str:
        return 'pcan' + match_object.group('channel')


class PCANTracePatternV13(PCANTracePatternV12):
    """
    Adds "Reserved" '-' column to 1.2
    1)      6357.213 1  Rx        0401 -  8    00 00 00 00 00 00 00 00

    >>> PCANTracePatternV13().match("  1)      6357.213 1  Rx        0401 -  8    00 00 00 00 00 00 00 00") #doctest: +ELLIPSIS
    <logreader.DataFrame object at ...>
    """
    pattern = re.compile(
        r'^\s*?\d+\)\s*?(?P<timestamp>\d+.\d+)\s+(?P<channel>[0-9])\s+(?P<type>\w+)\s+(?P<can_id>[0-9A-F]+)\s+-\s+(?P<dlc>[0-9])\s+(?P<can_data>RTR|[0-9A-F ]*)$')


class PCANTracePatternV20(PCANTracePatternV13):
    """
     1      1059.900 DT 0300 Rx 7 00 00 00 00 04 00 00

    >>> PCANTracePatternV20().match(" 1      1059.900 DT 0300 Rx 7 00 00 00 00 04 00 00") #doctest: +ELLIPSIS
    <logreader.DataFrame object at ...>
    """
    pattern = re.compile(
        r'^\s*?\d+?\s*?(?P<timestamp>\d+.\d+)\s+(?P<type>\w+)\s+(?P<can_id>[0-9A-F]+)\s+(?P<rxtx>\w+)\s+(?P<dlc>[0-9]+)(\s+(?P<can_data>[0-9A-F ]*))?$')

    def parse_channel(self, match_object: re.Match[str]) -> str:
        return 'pcanx'

    def parse_data(self, match_object: re.Match[str]) -> tuple[bytes, bool]:
        if match_object.group('type') == 'RR':
            is_remote_frame = True
            data = bytes(0)
        else:
            data, is_remote_frame = super().parse_data(match_object)
        return data, is_remote_frame


class PCANTracePatternV21(PCANTracePatternV20):
    """
    "Reserved" '-' and "Bus" to 2.0
     1      1059.900 DT 1 0300 Rx - 7 00 00 00 00 04 00 00

    >>> PCANTracePatternV21().match(" 1      1059.900 DT 1 0300 Rx - 7 00 00 00 00 04 00 00") #doctest: +ELLIPSIS
    <logreader.DataFrame object at ...>
    """
    pattern = re.compile(
        r'^\s*?\d+?\s*?(?P<timestamp>\d+.\d+)\s+(?P<type>\w+)\s+(?P<channel>[0-9])\s+(?P<can_id>[0-9A-F]+)\s+(?P<rxtx>.+)\s+-\s+(?P<dlc>[0-9]+)(\s+(?P<can_data>[0-9A-F ]*))?$')

    def parse_channel(self, match_object: re.Match[str]) -> str:
        return 'pcan' + match_object.group('channel')


class Parser:
    """A CAN log file parser.

    Automatically detects the format of the logfile by trying parser patterns
    until the first successful match.

    >>> with open('candump.log') as fd: #doctest: +SKIP
            for frame in cantools.logreader.Parser(fd):
                print(f'{frame.timestamp}: {frame.frame_id}')
    """

    def __init__(self, stream: io.TextIOBase | None = None, *, tz: TimezoneType = TZ_LOCAL) -> None:
        '''
        :param tz: The timezone which returned datetime objects should have when parsing `candump -l`, `candump -L` or `candump -ta`.
                   (It cannot be applied to `candump -tA` because the information is missing which timezone the time stamps have.
                    `candump -td`, `candump -tz` and PCAN trace do not have absolute time stamps but relative time stamps and return timedelta objects instead of datetime objects.)
                   With `tz=None` naive datetime objects representing the local system time will be returned.
                   With `tz=TZ_LOCAL` timezone-aware datetime objects in the local system time will be returned.
                   TZ_LOCAL should be preferred over None to avoid confusion with the naive datetime objects returned when parsing a `candump -tA` log.

        WARNING: Be careful with the datetime objects returned when parsing a `candump -tA` log. They are naive because the information is missing which timezone they represent.
        Usually naive datetime objects are assumed to represent the local system time but these objects are in the time zone where the log has been recorded.
        '''
        self.stream = stream
        self.pattern: BasePattern | None = None
        self.tz = tz

    def detect_pattern(self, line: str) -> BasePattern | None:
        for p in [CandumpDefaultPattern(), CandumpTimestampedPattern(self.tz), CandumpDefaultLogPattern(self.tz), CandumpAbsoluteLogPattern(), PCANTracePatternV21(), PCANTracePatternV20(), PCANTracePatternV13(), PCANTracePatternV12(), PCANTracePatternV11(), PCANTracePatternV10()]:
            mo = p.pattern.match(line)
            if mo:
                return p

        return None

    def parse(self, line: str) -> DataFrame | None:
        if self.pattern is None:
            self.pattern = self.detect_pattern(line)
        if self.pattern is None:
            return None
        return self.pattern.match(line)

    @overload
    def iterlines(self) -> Iterator[tuple[str, DataFrame]]:
        pass

    @overload
    def iterlines(self, keep_unknowns: Literal[True]) -> Iterator[tuple[str, DataFrame | None]]:
        pass

    def iterlines(self, keep_unknowns: bool = False) -> Iterator[tuple[str, DataFrame | None]]:
        """Returns an generator that yields (str, DataFrame) tuples with the
        raw log entry and a parsed log entry. If keep_unknowns=True, (str,
        None) tuples will be returned for log entries that couldn't be decoded.
        If keep_unknowns=False, non-parseable log entries is discarded.
        """
        if self.stream is None:
            return
        while True:
            nl = self.stream.readline()
            if nl == '':
                return
            nl = nl.strip('\r\n')
            frame = self.parse(nl)
            if frame:
                yield nl, frame
            elif keep_unknowns:
                yield nl, None
            else:
                continue

    def __iter__(self) -> Iterator[DataFrame]:
        """Returns DataFrame log entries. Non-parseable log entries is
        discarded."""
        for _, frame in self.iterlines():
            yield frame



# Store the specific DBC format properties of objects

from collections import OrderedDict


class DbcSpecifics:

    def __init__(self,
                 attributes=None,
                 attribute_definitions=None,
                 environment_variables=None,
                 value_tables=None,
                 attributes_rel=None,
                 attribute_definitions_rel=None):
        if attributes is None:
            attributes = OrderedDict()

        if attribute_definitions is None:
            attribute_definitions = OrderedDict()

        if environment_variables is None:
            environment_variables = OrderedDict()

        if value_tables is None:
            value_tables = OrderedDict()

        if attributes_rel is None:
            attributes_rel = OrderedDict()

        if attribute_definitions_rel is None:
            attribute_definitions_rel = OrderedDict()

        self._attributes = attributes
        self._attribute_definitions = attribute_definitions
        self._environment_variables = environment_variables
        self._value_tables = value_tables
        self._attributes_rel = attributes_rel
        self._attribute_definitions_rel = attribute_definitions_rel

    @property
    def attributes(self):
        """The DBC specific attributes of the parent object (database, node,
        message or signal) as a dictionary.

        """

        return self._attributes

    @attributes.setter
    def attributes(self, value):
        self._attributes = value

    @property
    def attribute_definitions(self):
        """The DBC specific attribute definitions as dictionary.

        """

        return self._attribute_definitions

    @property
    def value_tables(self):
        """An ordered dictionary of all value tables. Only valid for DBC
        specifiers on database level.

        """

        return self._value_tables

    @property
    def environment_variables(self):
        """An ordered dictionary of all environment variables. Only valid for
        DBC specifiers on database level.

        """

        return self._environment_variables

    @property
    def attributes_rel(self):
        """The DBC specific attribute rel as dictionary..

        """

        return self._attributes_rel

    @property
    def attribute_definitions_rel(self):
        """The DBC specific attribute definitions rel as dictionary.

        """

        return self._attribute_definitions_rel



import argparse
import os
import os.path

from .. import database
from ..database.can.c_source import camel_to_snake_case, generate


def _do_generate_c_source(args):
    dbase = database.load_file(args.infile,
                               encoding=args.encoding,
                               prune_choices=args.prune,
                               strict=not args.no_strict)

    if args.database_name is None:
        basename = os.path.basename(args.infile)
        database_name = os.path.splitext(basename)[0]
        database_name = camel_to_snake_case(database_name)
    else:
        database_name = args.database_name

    filename_h = database_name + '.h'
    filename_c = database_name + '.c'
    fuzzer_filename_c = database_name + '_fuzzer.c'
    fuzzer_filename_mk = database_name + '_fuzzer.mk'

    header, source, fuzzer_source, fuzzer_makefile = generate(
        dbase,
        database_name,
        filename_h,
        filename_c,
        fuzzer_filename_c,
        not args.no_floating_point_numbers,
        args.bit_fields,
        args.use_float,
        args.node,
        args.use_round)

    os.makedirs(args.output_directory, exist_ok=True)

    path_h = os.path.join(args.output_directory, filename_h)

    with open(path_h, 'w') as fout:
        fout.write(header)

    path_c = os.path.join(args.output_directory, filename_c)

    with open(path_c, 'w') as fout:
        fout.write(source)

    print(f'Successfully generated {path_h} and {path_c}.')

    if args.generate_fuzzer:
        fuzzer_path_c = os.path.join(args.output_directory, fuzzer_filename_c)

        with open(fuzzer_path_c, 'w') as fout:
            fout.write(fuzzer_source)

        fuzzer_path_mk = os.path.join(args.output_directory, fuzzer_filename_mk)

        with open(fuzzer_path_mk, 'w') as fout:
            fout.write(fuzzer_makefile)

        print(f'Successfully generated {fuzzer_path_c} and {fuzzer_path_mk}.')
        print()
        print(
            f'Run "make -f {fuzzer_filename_mk}" to build and run the fuzzer. Requires a')
        print('recent version of clang.')


def add_subparser(subparsers):
    generate_c_source_parser = subparsers.add_parser(
        'generate_c_source',
        description='Generate C source code from given database file.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    generate_c_source_parser.add_argument(
        '--database-name',
        help=('The database name.  Uses the stem of the input file name if not'
              ' specified.'))
    generate_c_source_parser.add_argument(
        '--no-floating-point-numbers',
        action='store_true',
        default=False,
        help='No floating point numbers in the generated code.')
    generate_c_source_parser.add_argument(
        '--bit-fields',
        action='store_true',
        help='Use bit fields to minimize struct sizes.')
    generate_c_source_parser.add_argument(
        '-e', '--encoding',
        help='File encoding.')
    generate_c_source_parser.add_argument(
        '--prune',
        action='store_true',
        help='Try to shorten the names of named signal choices.')
    generate_c_source_parser.add_argument(
        '--no-strict',
        action='store_true',
        help='Skip database consistency checks.')
    generate_c_source_parser.add_argument(
        '-f', '--generate-fuzzer',
        action='store_true',
        help='Also generate fuzzer source code.')
    generate_c_source_parser.add_argument(
        '-o', '--output-directory',
        default='.',
        help='Directory in which to write output files.')
    generate_c_source_parser.add_argument(
        '--use-float',
        action='store_true',
        default=False,
        help='Use float instead of double for floating point generation.')
    generate_c_source_parser.add_argument(
        '--use-round',
        action='store_true',
        default=False,
        help='Use rounding instead of truncation for floating point conversions.')
    generate_c_source_parser.add_argument(
        'infile',
        help='Input database file.')
    generate_c_source_parser.add_argument(
        '--node',
        help='Generate pack/unpack functions only for messages sent/received by the node.')
    generate_c_source_parser.set_defaults(func=_do_generate_c_source)





class AutosarSecOCProperties:
    """This class collects all attributes that are required to implement the
    AUTOSAR secure on-board communication (SecOC) specification.

    Be aware that the AUTOSAR SecOC specification does not cover the
    required cryptographic primitives themselves, just the
    "scaffolding" around them...
    """
    def __init__(self,
                 auth_algorithm_name: str | None,
                 freshness_algorithm_name: str | None,
                 payload_length: int | None,
                 data_id: int | None,
                 auth_tx_bit_length: int | None,
                 freshness_bit_length: int | None,
                 freshness_tx_bit_length: int | None,
                 ):

        self._auth_algorithm_name = auth_algorithm_name
        self._freshness_algorithm_name = freshness_algorithm_name

        self._payload_length = payload_length
        self._data_id = data_id

        self._freshness_bit_length = freshness_bit_length
        self._freshness_tx_bit_length = freshness_tx_bit_length
        self._auth_tx_bit_length = auth_tx_bit_length

    @property
    def freshness_algorithm_name(self) -> str | None:
        """The name of the algorithm used for verifying the freshness of a
        message.

        This can be used to prevent replay attacks. Note that the
        algorithms themselves are manufacturer-specific, i.e., AUTOSAR
        does not define *any* freshness schemes.
        """
        return self._freshness_algorithm_name

    @property
    def auth_algorithm_name(self) -> str | None:
        """The name of the algorithm used for authentication

        Note that the algorithms themselves are manufacturer-specific,
        i.e., AUTOSAR does not define *any* cryptographic schemes.
        """
        return self._auth_algorithm_name

    @property
    def payload_length(self) -> int | None:
        """Returns the number of bytes covered by the payload of the secured
        message

        (The full message length is the length of the payload plus the
        size of the security trailer.)
        """
        return self._payload_length

    @property
    def data_id(self) -> int | None:
        """The data ID required for authentication.

        Be aware that this is a different data ID than that required
        for End-To-End protection.
        """
        return self._data_id

    @property
    def freshness_bit_length(self) -> int | None:
        """The number of bits of the full freshness counter.
        """
        return self._freshness_bit_length

    @property
    def freshness_tx_bit_length(self) -> int | None:
        """The number of least-significant bits of the freshness counter that
        is send as part of the secured frame.

        This number is at most as large as the number of bits of
        freshness counter objects.

        """
        return self._freshness_tx_bit_length

    @property
    def auth_tx_bit_length(self) -> int | None:
        """The number of most significant bits of the authenticator object
        send as part of the secured frame

        This is at most the length of the authenicator.
        """
        return self._auth_tx_bit_length



# Internal diagnostics database.

from cantools.database.diagnostics.did import Did


class InternalDatabase:
    """Internal diagnostics database.

    """

    def __init__(self, dids: list[Did]) -> None:
        self.dids = dids



import argparse
import bisect
import curses
import queue
import re
import sys
import time
import warnings
from enum import Enum
from typing import Any

import can.cli
from argparse_addons import Integer  # type: ignore

from cantools.database.errors import DecodeError

from .. import database
from ..typechecking import SignalDictType
from .__utils__ import (
    format_multiplexed_name,
    format_signals,
    parse_additional_config,
)


class QuitError(Exception):
    pass

class MessageFormattingResult(Enum):
    Ok = 0
    UnknownMessage = 1
    DecodeError = 2

class Monitor(can.Listener):

    def __init__(self, stdscr: Any, args: argparse.Namespace):
        self._stdscr = stdscr
        print(f'Reading bus description file "{args.database}"...\r')
        self._dbase = database.load_file(args.database,
                                         encoding=args.encoding,
                                         frame_id_mask=args.frame_id_mask,
                                         prune_choices=args.prune,
                                         strict=not args.no_strict)
        self._single_line = args.single_line
        self._filtered_sorted_message_names: list[str] = []
        self._filter = args.filter_regex or ''
        self._filter_cursor_pos = 0
        self._compiled_filter = None
        self._formatted_messages: dict[str, list[str]] = {}
        self._raw_messages: dict[str, can.Message] = {}
        self._message_signals: dict[str, set[str]] = {}
        self._message_filtered_signals: dict[str, set[str]] = {}
        self._messages_with_error: set[str] = set()
        self._playing = True
        self._modified = True
        self._show_filter = False
        self._queue: queue.Queue = queue.Queue()
        self._nrows, self._ncols = stdscr.getmaxyx()
        self._received = 0
        self._discarded = 0
        self._errors = 0
        self._basetime: float | None = None
        self._page_first_row = 0

        if self._filter is not None:
            self.compile_filter()

        stdscr.keypad(True)
        stdscr.nodelay(True)
        curses.use_default_colors()
        curses.curs_set(False)
        curses.init_pair(1, curses.COLOR_BLACK, curses.COLOR_GREEN)
        curses.init_pair(2, curses.COLOR_BLACK, curses.COLOR_CYAN)
        curses.init_pair(3, curses.COLOR_CYAN, curses.COLOR_BLACK)

        if "bus_type" in args:
            # TODO: this branch can be removed once legacy args are removed
            bus = self.create_bus(args)
        else:
            bus = can.cli.create_bus_from_namespace(args)

        self._notifier = can.Notifier(bus, [self])

    def create_bus(self, args):
        kwargs = {}

        if args.bit_rate is not None:
            kwargs['bitrate'] = int(args.bit_rate)

        if args.fd:
            kwargs['fd'] = True

        if args.extra_args:
            kwargs.update(parse_additional_config(args.extra_args))

        try:
            return can.Bus(interface=args.bus_type,
                           channel=args.channel,
                           **kwargs)
        except Exception as exc:
            raise Exception(
                f"Failed to create CAN bus with bustype='{args.bus_type}' and "
                f"channel='{args.channel}'."
            ) from exc

    def run(self, max_num_keys_per_tick=-1):
        while True:
            try:
                self.tick(max_num_keys_per_tick)
            except QuitError:
                break

            time.sleep(0.05)

    def tick(self, max_num_keys=-1):
        modified = self.update()

        if modified:
            self.redraw()

        self.process_user_input(max_num_keys)

    def redraw(self):
        # Clear the screen.
        self._stdscr.erase()

        # Draw everything.
        self.draw_stats(0)
        self.draw_title(1)

        lines = []

        for name in self._filtered_sorted_message_names:
            for line in self._formatted_messages[name]:
                lines.append(line)

        # Only render the visible screen. We only have (self._nrows - 3)
        # available rows to draw on, due to the persistent TUI features that
        # are drawn:
        #
        # - line 0: stats
        # - line 1: title
        # - line (n - 1): menu
        num_actual_usable_rows = self._nrows - 2 - 1
        row = 2

        # make sure that we don't overshoot the last line of
        # content. this is a bit of a hack, because manipulation of
        # the controls is not supposed to happen within this method
        if len(lines) < self._page_first_row + num_actual_usable_rows:
            self._page_first_row = max(0, len(lines) - num_actual_usable_rows)

        for line in lines[self._page_first_row:self._page_first_row + num_actual_usable_rows]:
            self.addstr(row, 0, line)
            row += 1

        self.draw_menu(self._nrows - 1)

        # Refresh the screen.
        self._stdscr.refresh()

    def draw_stats(self, row):
        status_text = \
            f'Received: {self._received}, Discarded: {self._discarded}, Errors: {self._errors}'
        if self._filter:
            status_text += f', Filter: {self._filter}'
        self.addstr(row, 0, status_text)

    def draw_title(self, row):
        self.addstr_color(row,
                          0,
                          self.stretch('   TIMESTAMP  MESSAGE'),
                          curses.color_pair(1))

    def draw_menu(self, row):
        if self._show_filter:
            col = 0

            # text before cursor
            text = 'Filter regex: ' + self._filter[:self._filter_cursor_pos]
            self.addstr_color(row,
                              col,
                              text,
                              curses.color_pair(2))

            col = len(text)

            # cursor
            if self._filter_cursor_pos >= len(self._filter):
                c = " "
            else:
                c = self._filter[self._filter_cursor_pos]
            self.addstr_color(row,
                              col,
                              c,
                              curses.color_pair(3))
            col += 1

            # text after cursor
            text = self._filter[self._filter_cursor_pos + 1:]
            if len(text) > 0:
                self.addstr_color(row,
                                  col,
                                  text,
                                  curses.color_pair(2))
                col += len(text)

            # fill rest of line
            self.addstr_color(row,
                              col,
                              ' '*(self._ncols - col),
                              curses.color_pair(2))
        else:
            text = 'q: Quit, f: Filter, p: Play/Pause, r: Reset'

            self.addstr_color(row,
                              0,
                              self.stretch(text),
                              curses.color_pair(2))

    def addstr(self, row, col, text):
        try:
            self._stdscr.addstr(row, col, text)
        except curses.error:
            pass

    def addstr_color(self, row, col, text, color):
        try:
            self._stdscr.addstr(row, col, text, color)
        except curses.error:
            pass

    def stretch(self, text):
        return text + ' ' * (self._ncols - len(text))

    def process_user_input(self, max_num_keys=-1):
        while max_num_keys < 0 or max_num_keys > 0:
            max_num_keys -= 1
            try:
                key = self._stdscr.getkey()
            except curses.error:
                return

            if self._show_filter:
                self.process_user_input_filter(key)
            else:
                self.process_user_input_menu(key)

    def process_user_input_menu(self, key):
        if key == 'q':
            raise QuitError()
        elif key == 'p':
            self._playing = not self._playing
        elif key == 'r':
            self._playing = True
            self._filtered_sorted_message_names = []
            self._formatted_messages = {}
            self._received = 0
            self._discarded = 0
            self._basetime = None
            self._filter = ''
            self._compiled_filter = None
            self._modified = True
            self._page = 0

            while not self._queue.empty():
                self._queue.get()
        elif key in ['f', '/']:
            self._old_filter = self._filter
            self._show_filter = True
            self._filter_cursor_pos = len(self._filter)
            self._modified = True
            curses.curs_set(True)
        elif key == 'KEY_UP':
            self.line_up()
        elif key == 'KEY_DOWN':
            self.line_down()
        elif key == 'KEY_PPAGE':
            self.page_up()
        elif key == 'KEY_NPAGE':
            self.page_down()

    def line_down(self):
        # Increment line
        self._page_first_row += 1

        self._modified = True

    def line_up(self):
        # Decrement line
        if self._page_first_row > 0:
            self._page_first_row -= 1
        else:
            self._page_first_row = 0

        self._modified = True

    def page_up(self):
        num_actual_usable_rows = self._nrows - 2 - 1

        # Decrement page
        if self._page_first_row > num_actual_usable_rows:
            self._page_first_row -= num_actual_usable_rows
        else:
            self._page_first_row = 0

        self._modified = True

    def page_down(self):
        num_actual_usable_rows = self._nrows - 2 - 1

        # Increment page
        self._page_first_row += num_actual_usable_rows

        self._modified = True

    def compile_filter(self):
        try:
            self._compiled_filter = re.compile(self._filter, re.IGNORECASE)
        except (TypeError, re.error):
            self._compiled_filter = None

    def process_user_input_filter(self, key):
        if key == '\n':
            self._show_filter = False
            curses.curs_set(False)
        elif key == chr(27):
            # Escape
            self._show_filter = False
            self._filter = self._old_filter
            del self._old_filter
            curses.curs_set(False)
        elif key in ['KEY_BACKSPACE', '\b']:
            if self._filter_cursor_pos > 0:
                self._filter = \
                    self._filter[:self._filter_cursor_pos - 1] + \
                    self._filter[self._filter_cursor_pos:]
                self._filter_cursor_pos -= 1
        elif key == 'KEY_DC':
            # delete key
            if self._filter_cursor_pos < len(self._filter):
                self._filter = \
                    self._filter[:self._filter_cursor_pos] + \
                    self._filter[self._filter_cursor_pos + 1:]
        elif key == 'KEY_LEFT':
            if self._filter_cursor_pos > 0:
                self._filter_cursor_pos -= 1
        elif key == 'KEY_RIGHT':
            if self._filter_cursor_pos < len(self._filter):
                self._filter_cursor_pos += 1
        elif key == 'KEY_UP':
            self.line_up()
        elif key == 'KEY_DOWN':
            self.line_down()
        elif key == 'KEY_PPAGE':
            self.page_up()
        elif key == 'KEY_NPAGE':
            self.page_down()
        # we ignore keys with more than one character here. These
        # (mostly?) are control keys like KEY_UP, KEY_DOWN, etc.
        elif len(key) == 1:
            self._filter = \
                self._filter[:self._filter_cursor_pos] + \
                key + \
                self._filter[self._filter_cursor_pos:]
            self._filter_cursor_pos += 1

        self.compile_filter()

        # reformat all messages
        self._filtered_sorted_message_names.clear()
        self._message_signals.clear()
        self._formatted_messages.clear()
        for msg in self._raw_messages.values():
            self.try_update_message(msg)

        self._modified = True

    def try_update_message(self, raw_message: can.Message) -> MessageFormattingResult:
        if self._basetime is None:
            self._basetime = raw_message.timestamp

        data = raw_message.data
        timestamp = raw_message.timestamp - self._basetime

        try:
            message = self._dbase.get_message_by_frame_id(raw_message.arbitration_id, raw_message.is_extended_id) # type: ignore[union-attr]
        except KeyError:
            return MessageFormattingResult.UnknownMessage

        name = message.name
        try:
            if message.is_container:
                self._try_update_container(message, timestamp, data)
            else:
                if len(data) < message.length:
                    self._update_message_error(timestamp, name, data, f'{message.length - len(data)} bytes too short')
                    return MessageFormattingResult.DecodeError

                decoded_signals = message.decode_simple(bytes(data),
                    decode_choices=True,
                    allow_truncated=True,
                    allow_excess=True
                )

                name, formatted = self._format_message(timestamp, message, decoded_signals)
                self._update_formatted_message(name, formatted)
            self._raw_messages[name] = raw_message
            return MessageFormattingResult.Ok
        except DecodeError as e:
            # Discard the message in case of any decoding error, like we do when the
            # CAN message ID or length doesn't match what's specified in the DBC.
            self._update_message_error(timestamp, name, data, str(e))
            return MessageFormattingResult.DecodeError

    def _try_update_container(self, dbmsg, timestamp, data):
        decoded = dbmsg.decode(data, decode_containers=True)

        # handle the "table of contents" of the container message. To
        # avoid too much visual turmoil and the resulting usability issues,
        # we always put the contained messages on a single line
        contained_names = []
        for cmsg, _ in decoded:
            if isinstance(cmsg, int):
                tmp = dbmsg.get_contained_message_by_header_id(cmsg)
                cmsg_name = f'0x{cmsg:x}' if tmp is None else tmp.name
            else:
                cmsg_name = cmsg.name

            contained_names.append(cmsg_name)

        self._message_signals[dbmsg.name] = set(contained_names)
        self._update_formatted_message(dbmsg.name, self._format_lines(timestamp, dbmsg.name, contained_names))

        # handle the contained messages just as normal messages but
        # prefix their names with the name of the container followed
        # by '::'
        for cmsg, cdata in decoded:
            if isinstance(cmsg, int):
                tmp = dbmsg.get_contained_message_by_header_id(cmsg)
                cmsg_name = f'0x{cmsg:x}' if tmp is None else tmp.name
                full_name = f'{dbmsg.name} :: {cmsg_name}'

                if len(cdata) == 0:
                    cdata_str = f'<empty>'
                else:
                    cdata_str = f'0x{cdata.hex()}'

                formatted = self._format_lines(timestamp, full_name, [f'undecoded: {cdata_str}'])
            else:
                full_name, formatted = self._format_message(timestamp, cmsg, cdata, name_prefix=f'{dbmsg.name} :: ')
            self._update_formatted_message(full_name, formatted)

    def _format_message(self, timestamp: float, message: database.Message, decoded_signals: SignalDictType, name_prefix: str = '') -> tuple[str, list[str]]:
        name = message.name
        if message.is_multiplexed():
            name = format_multiplexed_name(message, decoded_signals)
        name = f'{name_prefix}{name}'

        filtered_signals = self._filter_signals(name, decoded_signals)
        formatted_signals = format_signals(message, filtered_signals)
        return name, self._format_lines(timestamp, name, formatted_signals)

    def _format_lines(self, timestamp: float, name: str, items: list[str], single_line: bool=False) -> list[str]:
        prefix = f'{timestamp:12.3f}  {name}('
        if self._single_line or single_line:
            formatted = [
                f'''{prefix}{', '.join(items)})'''
            ]
        else:
            formatted = [prefix]
            formatted += [f"{' ':<18}{line}{',' if index + 1 < len(items) else ''}" for index, line in enumerate(items)]
            formatted += [f"{' ':<14})"]
        return formatted


    def _filter_signals(self, name: str, signals: SignalDictType) -> SignalDictType:
        if name not in self._message_signals:
            self._message_signals[name] = self._message_filtered_signals[name] = set(signals.keys())
            self.insort_filtered(name)

        return {s: v for s, v in signals.items() if s in self._message_filtered_signals[name]}

    def _update_formatted_message(self, msg_name, formatted, is_error=False):
        old_formatted = self._formatted_messages.get(msg_name, [])

        # make sure never to decrease the number of lines occupied by
        # a message to avoid jittering
        if len(formatted) < len(old_formatted):
            formatted.extend(['']*(len(old_formatted) - len(formatted)))

        self._formatted_messages[msg_name] = formatted

        if is_error:
            self._messages_with_error.add(msg_name)
        else:
            self._messages_with_error.discard(msg_name)

        if msg_name not in self._filtered_sorted_message_names:
            self.insort_filtered(msg_name)

    def _update_message_error(self, timestamp, msg_name, data, error):
        formatted = self._format_lines(
            timestamp,
            msg_name,
            [f'undecoded, {error}: 0x{data.hex()}'],
            single_line=True
        )
        self._update_formatted_message(msg_name, formatted, is_error=True)

    def update_messages(self):
        modified = False

        try:
            while True:
                result = self.try_update_message(self._queue.get_nowait())
                self._received += 1
                if result == MessageFormattingResult.UnknownMessage:
                    self._discarded += 1
                elif result == MessageFormattingResult.DecodeError:
                    self._errors += 1
                modified = True
        except queue.Empty:
            pass

        return modified

    def update(self):
        if self._playing:
            modified = self.update_messages()
        else:
            modified = False

        if self._modified:
            self._modified = False
            modified = True

        if curses.is_term_resized(self._nrows, self._ncols):
            self._nrows, self._ncols = self._stdscr.getmaxyx()
            modified = True

        return modified

    def _signals_matching_filter(self, message_name: str) -> set[str]:
        all_signals = self._message_signals[message_name]

        # return all signals if:
        # - there is no filter
        # - message name matches the filter
        if self._compiled_filter is None or self._compiled_filter.search(message_name):
            return all_signals

        return {signal for signal in all_signals if self._compiled_filter.search(signal)}

    def _message_matches_filter(self, name: str) -> bool:
        # don't filter invalid messages as signals are unknown
        if name not in self._message_signals:
            return True

        matched_signals = self._signals_matching_filter(name)
        if matched_signals:
            self._message_filtered_signals[name] = matched_signals
        return bool(matched_signals)

    def insort_filtered(self, name):
        if name in self._filtered_sorted_message_names:
            return

        if name in self._messages_with_error or self._message_matches_filter(name):
            bisect.insort(self._filtered_sorted_message_names,
                          name)

    def on_message_received(self, msg):
        self._queue.put(msg)


def _do_monitor(args):
    def monitor(stdscr):
        Monitor(stdscr, args).run()

    try:
        curses.wrapper(monitor)
    except KeyboardInterrupt:
        pass


def _check_legacy_args() -> bool:
    """Return True if legacy CAN bus arguments were given."""
    args = sys.argv[1:]  # already a list of strings

    # New style explicitly supported -> no legacy mode
    if "-i" in args or "--interface" in args:
        return False

    # Special case: "-b" meaning change
    if "-b" in args:
        b_index = args.index("-b")
        # Legacy if the value after -b is non-numeric (old bus type behavior)
        if len(args) > b_index + 1 and not args[b_index + 1].isdigit():
            return True

    # Other explicit legacy indicators
    if any(arg in args for arg in ("--bus-type", "--bit-rate", "-B")):
        return True

    return False


def _warn_legacy_usage():
    """Emit a warning if legacy args were detected."""
    warnings.warn(
        "You are using legacy CAN bus arguments (-b, --bus-type, -B, etc.). "
        "These are deprecated and will be removed in a future release. "
        "Please use the new bus argument format via `--interface` and related options.",
        UserWarning,
        stacklevel=2
    )


def add_subparser(subparsers):
    monitor_parser = subparsers.add_parser(
        'monitor',
        description='Monitor CAN bus traffic in a text based user interface.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    monitor_parser.add_argument(
        '-s', '--single-line',
        action='store_true',
        help='Print the decoded message on a single line.')
    monitor_parser.add_argument(
        '-e', '--encoding',
        help='File encoding.')
    monitor_parser.add_argument(
        '-m', '--frame-id-mask',
        type=Integer(0),
        help=('Only compare selected frame id bits to find the message in the '
              'database. By default the received and database frame ids must '
              'be equal for a match.'))
    monitor_parser.add_argument(
        '--prune',
        action='store_true',
        help='Refrain from shortening the names of named signal values.')
    monitor_parser.add_argument(
        '--no-strict',
        action='store_true',
        help='Skip database consistency checks.')
    monitor_parser.add_argument(
        '-f', '--filter-regex',
        help='Use given filter regex.')

    if _check_legacy_args():
        _warn_legacy_usage()
        monitor_parser.add_argument(
            '-b', '--bus-type',
            default='socketcan',
            help='(Deprecated) Python CAN bus type.')
        monitor_parser.add_argument(
            '-c', '--channel',
            default='vcan0',
            help='(Deprecated) Python CAN bus channel.')
        monitor_parser.add_argument(
            '-B', '--bit-rate',
            help='(Deprecated) Python CAN bus bit rate.')
        monitor_parser.add_argument(
            '-f', '--fd',
            action='store_true',
            help='(Deprecated) Python CAN CAN-FD bus.')
        monitor_parser.add_argument(
            'extra_args',
            nargs=argparse.REMAINDER,
            help="(Deprecated) Remaining arguments will be used for the interface. "
                 "Example: `-c can0 -b socketcand --host=192.168.0.10 --port=29536` "
                 "is equivalent to `Bus('can0', 'socketcand', host='192.168.0.10', port=29536)`")
    else:
        can.cli.add_bus_arguments(monitor_parser, filter_arg=True, group_title="bus arguments (python-can)")

    monitor_parser.add_argument(
        'database',
        help='Database file.')
    monitor_parser.set_defaults(func=_do_monitor)



class AutosarDatabaseSpecifics:
    """This class collects the AUTOSAR specific information of a system

    """
    def __init__(self,
                 arxml_version):
        self._arxml_version = arxml_version

    @property
    def arxml_version(self):
        """The used version of ARXML file format

        Note that due to technical reasons we always return version
        "4.0.0" for AUTOSAR 4.X.
        """
        return self._arxml_version



import argparse

from .. import database


def _do_convert(args):
    dbase = database.load_file(args.infile,
                               encoding=args.encoding,
                               prune_choices=args.prune,
                               strict=not args.no_strict)

    database.dump_file(dbase,
                       args.outfile,
                       database_format=None,
                       encoding=args.encoding)


def add_subparser(subparsers):
    convert_parser = subparsers.add_parser(
        'convert',
        description='Convert given database from one format to another.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    convert_parser.add_argument(
        '-e', '--encoding',
        help='File encoding.')
    # the result of the convert operation should represent the
    # original file as closely as possible so -- in contrast to the
    # other subparsers -- we do not prune the names of signal values
    # by default
    convert_parser.add_argument(
        '--prune',
        action='store_true',
        help='Try to shorten the names of named signal choices.')
    convert_parser.add_argument(
        '--no-strict',
        action='store_true',
        help='Skip database consistency checks.')
    convert_parser.add_argument(
        'infile',
        help='Input database file.')
    convert_parser.add_argument(
        'outfile',
        help='Output database file.')
    convert_parser.set_defaults(func=_do_convert)



import logging
from collections import OrderedDict
from typing import (
    Any,
    TextIO,
)

from ...typechecking import DecodeResultType, EncodeInputType, StringPathLike
from ..errors import DecodeError
from ..utils import (
    SORT_SIGNALS_DEFAULT,
    sort_signals_by_start_bit,
    type_sort_attributes,
    type_sort_choices,
    type_sort_signals,
)
from .bus import Bus
from .formats import arxml, dbc, kcd, sym
from .formats.arxml import AutosarDatabaseSpecifics
from .formats.dbc import DbcSpecifics
from .internal_database import InternalDatabase
from .message import Message
from .node import Node

LOGGER = logging.getLogger(__name__)


class Database:
    """This class contains all messages, signals and definitions of a CAN
    network.

    The factory functions :func:`load()<cantools.database.load()>`,
    :func:`load_file()<cantools.database.load_file()>` and
    :func:`load_string()<cantools.database.load_string()>` returns
    instances of this class.

    If `strict` is ``True`` an exception is raised if any signals are
    overlapping or if they don't fit in their message.

    By default signals are sorted by their start bit when their Message object is created.
    If you don't want them to be sorted pass `sort_signals = None`.
    If you want the signals to be sorted in another way pass something like
    `sort_signals = lambda signals: list(sorted(signals, key=lambda sig: sig.name))`
    """

    def __init__(self,
                 messages: list[Message] | None = None,
                 nodes: list[Node] | None = None,
                 buses: list[Bus] | None = None,
                 version: str | None = None,
                 dbc_specifics: DbcSpecifics | None = None,
                 autosar_specifics: AutosarDatabaseSpecifics | None = None,
                 frame_id_mask: int | None = None,
                 strict: bool = True,
                 sort_signals: type_sort_signals = sort_signals_by_start_bit,
                 ) -> None:
        self._messages = messages or []
        self._nodes = nodes or []
        self._buses = buses or []
        self._name_to_message: dict[str, Message] = {}
        self._frame_id_to_message: dict[int, Message] = {}
        self._version = version
        self._dbc = dbc_specifics
        self._autosar = autosar_specifics

        if frame_id_mask is None:
            frame_id_mask = 0xffffffff

        self._frame_id_mask = frame_id_mask
        self._strict = strict
        self._sort_signals = sort_signals
        self.refresh()

    @property
    def messages(self) -> list[Message]:
        """A list of messages in the database.

        Use :meth:`.get_message_by_frame_id()` or
        :meth:`.get_message_by_name()` to find a message by its frame
        id or name.

        """

        return self._messages

    @property
    def nodes(self) -> list[Node]:
        """A list of nodes in the database.

        """

        return self._nodes

    @property
    def buses(self) -> list[Bus]:
        """A list of CAN buses in the database.

        """

        return self._buses

    @property
    def version(self) -> str | None:
        """The database version, or ``None`` if unavailable.

        """

        return self._version

    @version.setter
    def version(self, value: str | None) -> None:
        self._version = value

    @property
    def dbc(self) -> DbcSpecifics | None:
        """An object containing dbc specific properties like e.g. attributes.

        """

        return self._dbc

    @dbc.setter
    def dbc(self, value: DbcSpecifics | None) -> None:
        self._dbc = value

    @property
    def autosar(self) -> AutosarDatabaseSpecifics | None:
        """An object containing AUTOSAR specific properties like e.g. attributes.

        """

        return self._autosar

    @autosar.setter
    def autosar(self, value: AutosarDatabaseSpecifics | None) -> None:
        self._autosar = value


    def is_similar(self,
                   other: "Database",
                   *,
                   tolerance: float = 1e-12,
                   include_format_specifics: bool = True) -> bool:
        """Compare two database objects inexactly

        This means that small discrepanceies stemming from
        e.g. rounding errors are ignored.
        """
        return not self._differences(other,
                                     tolerance=tolerance,
                                     include_format_specifics=include_format_specifics)

    def _differences(self,
                     other: "Database",
                     *,
                     tolerance: float = 1e-12,
                     include_format_specifics: bool = True) -> list[dict[str, str]]:
        """Return a list of differences between this Database and `other`.

        Each difference is a dict with keys: ``path``, ``self``, ``other``, ``reason``.
        """
        diffs: list[dict[str, str]] = []
        seen: set[tuple[int, int]] = set()

        def path_to_str(path_parts: list[str]) -> str:
            if not path_parts:
                return '<root>'
            return ''.join(path_parts)

        def add_diff(path_parts: list[str], a_val: Any, b_val: Any, reason: str) -> None:
            diffs.append({
                'path': path_to_str(path_parts),
                'self': str(a_val),
                'other': str(b_val),
                'reason': reason,
            })

        def compare(a: Any, b: Any, path_parts: list[str]) -> None:
            pair_id = (id(a), id(b))
            if pair_id in seen:
                return
            seen.add(pair_id)

            if type(a) is not type(b):
                add_diff(path_parts, type(a), type(b), 'type-mismatch')
                return

            if callable(a) or a is None:
                # nothing to compare
                return

            if isinstance(a, (int, str, set, bool)):
                if a != b:
                    add_diff(path_parts, a, b, 'value-mismatch')
                return

            if isinstance(a, float):
                if abs(a) > 1:
                    if abs(1.0 - b / a) > tolerance:
                        add_diff(path_parts, a, b, 'float-rel-diff')
                elif abs(b - a) > tolerance:
                    add_diff(path_parts, a, b, 'float-abs-diff')
                return

            if isinstance(a, (list, tuple)):
                if len(a) != len(b):
                    add_diff(path_parts, len(a), len(b), 'length-mismatch')
                min_len = min(len(a), len(b))
                for i in range(min_len):
                    compare(a[i], b[i], [*path_parts, f'[{i}]'])
                return

            if isinstance(a, (dict, OrderedDict)):
                keys_a = set(a.keys())
                keys_b = set(b.keys())
                for k in sorted(keys_a - keys_b, key=str):
                    add_diff([*path_parts, f'[{k!r}]'], a[k], None, 'key-only-in-a')
                for k in sorted(keys_b - keys_a, key=str):
                    add_diff([*path_parts, f'[{k!r}]'], None, b[k], 'key-only-in-b')
                for k in sorted(keys_a & keys_b, key=str):
                    compare(a[k], b[k], [*path_parts, f'[{k!r}]'])
                return

            # get attributes
            a_names = dir(a)
            b_names = dir(b)

            if not include_format_specifics:
                for x in ('dbc', 'autosar'):
                    if x in a_names:
                        a_names.remove(x)
                    if x in b_names:
                        b_names.remove(x)

            if a_names != b_names:
                add_diff(path_parts, sorted(a_names), sorted(b_names), 'attrib-names-mismatch')
                return

            for name in a_names:
                if name == 'messages' and hasattr(a, '_frame_id_to_message'):
                    # compare messages independent of order
                    compare(a._frame_id_to_message, b._frame_id_to_message, [*path_parts, "._frame_id_to_message"])
                    continue

                if name == 'signals' and hasattr(a, '_signal_dict'):
                    # compare messages independent of order
                    compare(a._signal_dict, b._signal_dict, [*path_parts, "._signal_dict"])
                    continue

                if name.startswith('_'):
                    # skip private attributes
                    continue

                a_attr = getattr(a, name)
                b_attr = getattr(b, name)
                compare(a_attr, b_attr, [*path_parts, f'.{name}'])

        # compare root
        compare(self, other, [])

        return diffs

    def add_arxml(self, fp: TextIO) -> None:
        """Read and parse ARXML data from given file-like object and add the
        parsed data to the database.

        """

        self.add_arxml_string(fp.read())

    def add_arxml_file(self,
                       filename: StringPathLike,
                       encoding: str = 'utf-8') -> None:
        """Open, read and parse ARXML data from given file and add the parsed
        data to the database.

        `encoding` specifies the file encoding.

        """

        with open(filename, encoding=encoding, errors='replace') as fin:
            self.add_arxml(fin)

    def add_arxml_string(self, string: str) -> None:
        """Parse given ARXML data string and add the parsed data to the
        database.

        """

        database = arxml.load_string(string, self._strict, sort_signals=self._sort_signals)

        self._messages += database.messages
        self._nodes = database.nodes
        self._buses = database.buses
        self._version = database.version
        self._dbc = database.dbc
        self._autosar = database.autosar
        self.refresh()

    def add_dbc(self, fp: TextIO) -> None:
        """Read and parse DBC data from given file-like object and add the
        parsed data to the database.

        >>> db = cantools.database.Database()
        >>> with open ('foo.dbc', 'r') as fin:
        ...     db.add_dbc(fin)

        """

        self.add_dbc_string(fp.read())

    def add_dbc_file(self,
                     filename: StringPathLike,
                     encoding: str = 'cp1252') -> None:
        """Open, read and parse DBC data from given file and add the parsed
        data to the database.

        `encoding` specifies the file encoding.

        >>> db = cantools.database.Database()
        >>> db.add_dbc_file('foo.dbc')

        """

        with open(filename, encoding=encoding, errors='replace') as fin:
            self.add_dbc(fin)

    def add_dbc_string(self, string: str) -> None:
        """Parse given DBC data string and add the parsed data to the
        database.

        >>> db = cantools.database.Database()
        >>> with open ('foo.dbc', 'r') as fin:
        ...     db.add_dbc_string(fin.read())

        """

        database = dbc.load_string(string, self._strict, sort_signals=self._sort_signals)

        self._messages += database.messages
        self._nodes = database.nodes
        self._buses = database.buses
        self._version = database.version
        self._dbc = database.dbc
        self.refresh()

    def add_kcd(self, fp: TextIO) -> None:
        """Read and parse KCD data from given file-like object and add the
        parsed data to the database.

        """

        self.add_kcd_string(fp.read())

    def add_kcd_file(self,
                     filename: StringPathLike,
                     encoding: str = 'utf-8') -> None:
        """Open, read and parse KCD data from given file and add the parsed
        data to the database.

        `encoding` specifies the file encoding.

        """

        with open(filename, encoding=encoding, errors='replace') as fin:
            self.add_kcd(fin)

    def add_kcd_string(self, string: str) -> None:
        """Parse given KCD data string and add the parsed data to the
        database.

        """

        database = kcd.load_string(string, self._strict, sort_signals=self._sort_signals)

        self._messages += database.messages
        self._nodes = database.nodes
        self._buses = database.buses
        self._version = database.version
        self._dbc = database.dbc
        self.refresh()

    def add_sym(self, fp: TextIO) -> None:
        """Read and parse SYM data from given file-like object and add the
        parsed data to the database.

        """

        self.add_sym_string(fp.read())

    def add_sym_file(self,
                     filename: StringPathLike,
                     encoding: str = 'utf-8') -> None:
        """Open, read and parse SYM data from given file and add the parsed
        data to the database.

        `encoding` specifies the file encoding.

        """

        with open(filename, encoding=encoding, errors='replace') as fin:
            self.add_sym(fin)

    def add_sym_string(self, string: str) -> None:
        """Parse given SYM data string and add the parsed data to the
        database.

        """

        database = sym.load_string(string, self._strict, sort_signals=self._sort_signals)

        self._messages += database.messages
        self._nodes = database.nodes
        self._buses = database.buses
        self._version = database.version
        self._dbc = database.dbc
        self.refresh()

    def _add_message(self, message: Message) -> None:
        """Add given message to the database.

        """

        if message.name in self._name_to_message:
            LOGGER.warning("Overwriting message '%s' with '%s' in the "
                           "name to message dictionary.",
                           self._name_to_message[message.name].name,
                           message.name)

        masked_frame_id = (message.frame_id & self._frame_id_mask)
        if message.is_extended_frame:
            masked_frame_id |= 0x80000000

        if masked_frame_id in self._frame_id_to_message:
            LOGGER.warning(
                "Overwriting message '%s' with '%s' in the frame id to message "
                "dictionary because they have identical masked frame ids 0x%x.",
                self._frame_id_to_message[masked_frame_id].name,
                message.name,
                masked_frame_id)

        self._name_to_message[message.name] = message
        self._frame_id_to_message[masked_frame_id] = message

    def as_dbc_string(self, *,
                      sort_signals:type_sort_signals=SORT_SIGNALS_DEFAULT,
                      sort_attribute_signals:type_sort_signals=SORT_SIGNALS_DEFAULT,
                      sort_attributes:type_sort_attributes=None,
                      sort_choices:type_sort_choices=None,
                      shorten_long_names:bool=True) -> str:
        """Return the database as a string formatted as a DBC file.
           sort_signals defines how to sort signals in message definitions
           sort_attribute_signals defines how to sort signals in metadata -
              comments, value table definitions and attributes

        """
        if not self._sort_signals and sort_signals == SORT_SIGNALS_DEFAULT:
            sort_signals = None

        return dbc.dump_string(InternalDatabase(self._messages,
                                                self._nodes,
                                                self._buses,
                                                self._version,
                                                self._dbc),
                               sort_signals=sort_signals,
                               sort_attribute_signals=sort_attribute_signals,
                               sort_attributes=sort_attributes,
                               sort_choices=sort_choices,
                               shorten_long_names=shorten_long_names)

    def as_kcd_string(self, *, sort_signals:type_sort_signals=SORT_SIGNALS_DEFAULT) -> str:
        """Return the database as a string formatted as a KCD file.

        """
        if not self._sort_signals and sort_signals == SORT_SIGNALS_DEFAULT:
            sort_signals = None

        return kcd.dump_string(InternalDatabase(self._messages,
                                                self._nodes,
                                                self._buses,
                                                self._version,
                                                self._dbc),
                               sort_signals=sort_signals)

    def as_sym_string(self, *, sort_signals:type_sort_signals=SORT_SIGNALS_DEFAULT) -> str:
        """Return the database as a string formatted as a SYM file.

        """
        if not self._sort_signals and sort_signals == SORT_SIGNALS_DEFAULT:
            sort_signals = None

        return sym.dump_string(InternalDatabase(self._messages,
                                                self._nodes,
                                                self._buses,
                                                self._version,
                                                self._dbc),
                               sort_signals=sort_signals)

    def get_message_by_name(self, name: str) -> Message:
        """Find the message object for given name `name`.

        """

        return self._name_to_message[name]

    def get_message_by_frame_id(self, frame_id: int, force_extended_id: bool = False) -> Message:
        """Find the message object for given frame id `frame_id`.

        """

        if force_extended_id or frame_id > 0x7FF:
            frame_id |= 0x80000000

        return self._frame_id_to_message[frame_id & (0x80000000 | self._frame_id_mask)]

    def get_node_by_name(self, name: str) -> Node:
        """Find the node object for given name `name`.

        """

        for node in self._nodes:
            if node.name == name:
                return node

        raise KeyError(name)

    def get_bus_by_name(self, name: str) -> Bus:
        """Find the bus object for given name `name`.

        """

        for bus in self._buses:
            if bus.name == name:
                return bus

        raise KeyError(name)

    def encode_message(self,
                       frame_id_or_name: int | str,
                       data: EncodeInputType,
                       scaling: bool = True,
                       padding: bool = False,
                       strict: bool = True,
                       force_extended_id: bool = False,
                       ) -> bytes:
        """Encode given signal data `data` as a message of given frame id or
        name `frame_id_or_name`. For regular Messages, `data` is a
        dictionary of signal name-value entries, for container
        messages it is a list of (ContainedMessageOrMessageName,
        ContainedMessageSignals) tuples.

        If `scaling` is ``False`` no scaling of signals is performed.

        If `padding` is ``True`` unused bits are encoded as 1.

        If `strict` is ``True`` all signal values must be within their
        allowed ranges, or an exception is raised.

        >>> db.encode_message(158, {'Bar': 1, 'Fum': 5.0})
        b'\\x01\\x45\\x23\\x00\\x11'
        >>> db.encode_message('Foo', {'Bar': 1, 'Fum': 5.0})
        b'\\x01\\x45\\x23\\x00\\x11'

        """

        if isinstance(frame_id_or_name, int):
            if force_extended_id or frame_id_or_name > 0x7FF:
                frame_id_or_name |= 0x80000000
            message = self._frame_id_to_message[frame_id_or_name]
        elif isinstance(frame_id_or_name, str):
            message = self._name_to_message[frame_id_or_name]
        else:
            raise ValueError(f"Invalid frame_id_or_name '{frame_id_or_name}'")

        return message.encode(data, scaling, padding, strict)

    def decode_message(self,
                       frame_id_or_name: int | str,
                       data: bytes,
                       decode_choices: bool = True,
                       scaling: bool = True,
                       decode_containers: bool = False,
                       allow_truncated:  bool = False,
                       force_extended_id: bool = False,
                       ) \
        -> DecodeResultType:

        """Decode given signal data `data` as a message of given frame id or
        name `frame_id_or_name`. Returns a dictionary of signal
        name-value entries.

        If `decode_choices` is ``False`` scaled values are not
        converted to choice strings (if available).

        If `scaling` is ``False`` no scaling of signals is performed.

        >>> db.decode_message(158, b'\\x01\\x45\\x23\\x00\\x11')
        {'Bar': 1, 'Fum': 5.0}
        >>> db.decode_message('Foo', b'\\x01\\x45\\x23\\x00\\x11')
        {'Bar': 1, 'Fum': 5.0}

        If `decode_containers` is ``True``, container frames are
        decoded. The reason why this needs to be explicitly enabled is
        that decoding container frames returns a list of ``(Message,
        SignalsDict)`` tuples which will cause code that does not
        expect this to misbehave. Trying to decode a container message
        with `decode_containers` set to ``False`` will raise a
        `DecodeError`.
        """

        if isinstance(frame_id_or_name, int):
            if force_extended_id or frame_id_or_name > 0x7FF:
                frame_id_or_name |= 0x80000000
            message = self._frame_id_to_message[frame_id_or_name]
        elif isinstance(frame_id_or_name, str):
            message = self._name_to_message[frame_id_or_name]
        else:
            raise ValueError(f"Invalid frame_id_or_name '{frame_id_or_name}'")

        if message.is_container:
            if decode_containers:
                return message.decode(data,
                                      decode_choices,
                                      scaling,
                                      decode_containers=True,
                                      allow_truncated=allow_truncated)
            else:
                raise DecodeError(f'Message "{message.name}" is a container '
                                  f'message, but decoding such messages has '
                                  f'not been enabled!')

        return message.decode(data,
                              decode_choices,
                              scaling,
                              allow_truncated=allow_truncated)

    def refresh(self) -> None:
        """Refresh the internal database state.

        This method must be called after modifying any message in the
        database to refresh the internal lookup tables used when
        encoding and decoding messages.

        """

        self._name_to_message = {}
        self._frame_id_to_message = {}

        for message in self._messages:
            message.refresh(self._strict)
            self._add_message(message)

    def __repr__(self) -> str:
        lines = [f"version('{self._version}')", '']

        if self._nodes:
            for node in self._nodes:
                lines.append(repr(node))

            lines.append('')

        for message in self._messages:
            lines.append(repr(message))

            for signal in message.signals:
                lines.append('  ' + repr(signal))

            lines.append('')

        return '\n'.join(lines)



# Load and dump a diagnostics database in CDD format.
import logging
from collections import OrderedDict
from xml.etree import ElementTree

from cantools.typechecking import ByteOrder, Choices

from ...conversion import BaseConversion
from ...errors import ParseError
from ...utils import cdd_offset_to_dbc_start_bit
from ..data import Data
from ..did import Did
from ..internal_database import InternalDatabase

LOGGER = logging.getLogger(__name__)


class DataType:

    def __init__(self,
                 name: str,
                 id_: str,
                 bit_length: int,
                 encoding: str | None,
                 minimum: int | None,
                 maximum: int | None,
                 choices: Choices | None,
                 byte_order: ByteOrder,
                 unit: str | None,
                 factor: float,
                 offset: float) -> None:
        self.name = name
        self.id_ = id_
        self.bit_length = bit_length
        self.encoding = encoding
        self.minimum = minimum
        self.maximum = maximum
        self.choices = choices
        self.byte_order = byte_order
        self.unit = unit
        self.factor = factor
        self.offset = offset


def _load_choices(data_type: ElementTree.Element) -> Choices | None:
    choices: Choices = OrderedDict()

    for choice in data_type.findall('TEXTMAP'):
        start = int(choice.attrib['s'].strip('()'))
        end = int(choice.attrib['e'].strip('()'))

        if start == end:
            choice_text = choice.findtext('TEXT/TUV[1]')
            if choice_text is None:
                raise ValueError(f"Could not find name in TUV!")
            choices[start] = choice_text

    if not choices:
        return None

    return choices


def _load_data_types(ecu_doc: ElementTree.Element | None) -> dict[str, DataType]:
    """Load all data types found in given ECU doc element.

    """

    data_types: dict[str, DataType] = {}

    if ecu_doc is None:
        return data_types

    types = ecu_doc.findall('DATATYPES/*')

    for data_type in types:
        # Default values.
        byte_order: ByteOrder = 'big_endian'
        unit = None
        factor: float = 1
        offset: float = 0
        bit_length = 0
        encoding = None
        minimum = None
        maximum = None

        # Name and id.
        type_name = data_type.findtext('NAME/TUV[1]')
        if type_name is None:
            raise ValueError(f"Could not find name in DATATYPE IDENT with id={data_type.attrib.get('id')}!")
        type_id = data_type.attrib['id']

        # Load from C-type element.
        ctype = data_type.find('CVALUETYPE')
        if ctype is not None:
            for key, value in ctype.attrib.items():
                if key == 'bl':
                    bit_length = int(value)
                elif key == 'enc':
                    encoding = value
                elif key == 'minsz':
                    minimum = int(value)
                elif key == 'maxsz':
                    maximum = int(value)
                else:
                    LOGGER.debug("Ignoring unsupported attribute '%s'.", key)

            if ctype.attrib['bo'] == '21':
                byte_order = 'big_endian'
            elif ctype.attrib['bo'] == '12':
                byte_order = 'little_endian'
            else:
                raise ParseError(f"Unknown byte order code: {ctype.attrib['bo']}")

        # Load from P-type element.
        unit = data_type.findtext('PVALUETYPE/UNIT')

        # Choices, scale and offset.
        choices = _load_choices(data_type)

        # Slope and offset.
        comp = data_type.find('COMP')

        if comp is not None:
            factor = float(comp.attrib['f'])
            offset = float(comp.attrib['o'])

        if bit_length == 0:
            raise RuntimeError("CVALUETYPE element cannot have bit length 0!")

        data_types[type_id] = DataType(type_name,
                                       type_id,
                                       bit_length,
                                       encoding,
                                       minimum,
                                       maximum,
                                       choices,
                                       byte_order,
                                       unit,
                                       factor,
                                       offset)

    return data_types


def _load_data_element(data: ElementTree.Element, offset: int, data_types: dict[str, DataType]) -> Data:
    """Load given signal element and return a signal object.

    """

    data_type = data_types[data.attrib['dtref']]

    # Map CDD/c-style field offset to the DBC/can.Signal.start bit numbering
    # convention for compatibility with can.Signal objects and the shared codec
    # infrastructure.
    #
    dbc_start_bitnum = cdd_offset_to_dbc_start_bit(offset, data_type.bit_length, data_type.byte_order)

    conversion = BaseConversion.factory(
        scale=data_type.factor,
        offset=data_type.offset,
        choices=data_type.choices,
        is_float=False
    )

    name = data.findtext('QUAL')
    if name is None:
        raise ValueError(f"Could not get QUAL text in data with id={data.attrib.get('id')}!")

    return Data(name=name,
                start=dbc_start_bitnum,
                length=data_type.bit_length,
                byte_order=data_type.byte_order,
                conversion=conversion,
                minimum=data_type.minimum,
                maximum=data_type.maximum,
                unit=data_type.unit)


def _load_did_element(did: ElementTree.Element, data_types: dict[str, DataType], did_data_lib: dict[str, ElementTree.Element]) -> Did:
    """Load given DID element and return a did object.

    """

    offset = 0
    datas: list[Data] = []
    data_objs = did.findall('SIMPLECOMPCONT/DATAOBJ')
    data_objs += did.findall('SIMPLECOMPCONT/UNION/STRUCT/DATAOBJ')
    did_data_refs = did.findall('SIMPLECOMPCONT/DIDDATAREF')

    for data_ref in did_data_refs:
        try:
            data_objs += did_data_lib[data_ref.attrib['didRef']].findall('STRUCTURE/DATAOBJ')
        except KeyError:
            pass

    for data_obj in data_objs:
        data = _load_data_element(data_obj,
                                  offset,
                                  data_types)

        if data:
            datas.append(data)
            offset += data.length or 0

    static_value = did.find('STATICVALUE')
    if static_value is None:
        raise KeyError(f"Could not find STATICVALUE element in DID with id={did.attrib.get('id')}!")
    identifier = int(static_value.attrib['v'])
    name = did.findtext('QUAL')
    if name is None:
        raise ValueError(f"Could not get QUAL text in DID with id={did.attrib.get('id')}!")
    length = (offset + 7) // 8

    return Did(identifier=identifier,
               name=name,
               length=length,
               datas=datas)


def _load_did_data_refs(ecu_doc: ElementTree.Element | None) -> dict[str, ElementTree.Element]:
    """Load DID data references from given ECU doc element.

    """
    if ecu_doc is None:
        return {}

    dids = ecu_doc.find('DIDS')

    if dids is None:
        return {}
    else:
        return {did.attrib['id']: did for did in dids.findall('DID')}


def load_string(string: str) -> InternalDatabase:
    """Parse given CDD format string.

    """

    root = ElementTree.fromstring(string)
    ecu_doc = root.find('ECUDOC')
    if ecu_doc is None:
        raise KeyError("Could not find ECUDOC root element!")
    data_types = _load_data_types(ecu_doc)
    did_data_lib = _load_did_data_refs(ecu_doc)
    var = ecu_doc.findall('ECU')[0].find('VAR')
    if var is None:
        raise KeyError(f"Could not find VAR element in ECU with id={ecu_doc.findall('ECU')[0].attrib.get('id')}!")
    dids: list[Did] = []

    for diag_class in var.findall('DIAGCLASS'):
        for diag_inst in diag_class.findall('DIAGINST'):
            did = _load_did_element(diag_inst,
                                    data_types,
                                    did_data_lib)
            dids.append(did)

    return InternalDatabase(dids)



# Load and dump a CAN database in DBC format.

import re
from collections import OrderedDict, defaultdict
from copy import deepcopy
from decimal import Decimal

import textparser  # type: ignore
from textparser import (
    Any,
    AnyUntil,
    DelimitedList,
    OneOrMore,
    OneOrMoreDict,
    Optional,
    Sequence,
    Token,
    TokenizeError,
    ZeroOrMore,
    choice,
    tokenize_init,
)

from ...conversion import BaseConversion
from ...namedsignalvalue import NamedSignalValue
from ...utils import (
    SORT_SIGNALS_DEFAULT,
    sort_signals_by_start_bit,
    sort_signals_by_start_bit_reversed,
    type_sort_attributes,
    type_sort_choices,
    type_sort_signals,
)
from ..attribute import Attribute
from ..attribute_definition import AttributeDefinition
from ..bus import Bus
from ..environment_variable import EnvironmentVariable
from ..internal_database import InternalDatabase
from ..message import Message
from ..node import Node
from ..signal import Signal
from ..signal_group import SignalGroup
from .dbc_specifics import DbcSpecifics
from .utils import num

DBC_FMT = (
    'VERSION "{version}"\r\n'
    '\r\n'
    '\r\n'
    'NS_ : \r\n'
    '\tNS_DESC_\r\n'
    '\tCM_\r\n'
    '\tBA_DEF_\r\n'
    '\tBA_\r\n'
    '\tVAL_\r\n'
    '\tCAT_DEF_\r\n'
    '\tCAT_\r\n'
    '\tFILTER\r\n'
    '\tBA_DEF_DEF_\r\n'
    '\tEV_DATA_\r\n'
    '\tENVVAR_DATA_\r\n'
    '\tSGTYPE_\r\n'
    '\tSGTYPE_VAL_\r\n'
    '\tBA_DEF_SGTYPE_\r\n'
    '\tBA_SGTYPE_\r\n'
    '\tSIG_TYPE_REF_\r\n'
    '\tVAL_TABLE_\r\n'
    '\tSIG_GROUP_\r\n'
    '\tSIG_VALTYPE_\r\n'
    '\tSIGTYPE_VALTYPE_\r\n'
    '\tBO_TX_BU_\r\n'
    '\tBA_DEF_REL_\r\n'
    '\tBA_REL_\r\n'
    '\tBA_DEF_DEF_REL_\r\n'
    '\tBU_SG_REL_\r\n'
    '\tBU_EV_REL_\r\n'
    '\tBU_BO_REL_\r\n'
    '\tSG_MUL_VAL_\r\n'
    '\r\n'
    'BS_:\r\n'
    '\r\n'
    'BU_: {bu}\r\n'
    '{val_table}'
    '\r\n'
    '\r\n'
    '{bo}\r\n'
    '\r\n'
    '{bo_tx_bu}\r\n'
    '{ev}\r\n'
    '\r\n'
    '{cm}\r\n'
    '{ba_def}\r\n'
    '{ba_def_rel}'
    '{ba_def_def}\r\n'
    '{ba_def_def_rel}'
    '{ba}\r\n'
    '{ba_rel}'
    '{val}\r\n'
    '{signal_types}\r\n'
    '{sig_group}\r\n'
    '{sig_mux_values}\r\n'
)


# Signal types.
SIGNAL_TYPE_FLOAT = 1
SIGNAL_TYPE_DOUBLE = 2

FLOAT_SIGNAL_TYPES = [
    SIGNAL_TYPE_FLOAT,
    SIGNAL_TYPE_DOUBLE
]

FLOAT_LENGTH_TO_SIGNAL_TYPE = {
    32: SIGNAL_TYPE_FLOAT,
    64: SIGNAL_TYPE_DOUBLE
}

ATTRIBUTE_DEFINITION_LONG_ENVVAR_NAME = AttributeDefinition(
    'SystemEnvVarLongSymbol',
    default_value='',
    kind='EV_',
    type_name='STRING')
ATTRIBUTE_DEFINITION_LONG_NODE_NAME = AttributeDefinition(
    'SystemNodeLongSymbol',
    default_value='',
    kind='BU_',
    type_name='STRING')

ATTRIBUTE_DEFINITION_LONG_MESSAGE_NAME = AttributeDefinition(
    'SystemMessageLongSymbol',
    default_value='',
    kind='BO_',
    type_name='STRING')

ATTRIBUTE_DEFINITION_LONG_SIGNAL_NAME = AttributeDefinition(
    'SystemSignalLongSymbol',
    default_value='',
    kind='SG_',
    type_name='STRING')

ATTRIBUTE_DEFINITION_VFRAMEFORMAT = AttributeDefinition(
    name='VFrameFormat',
    default_value='StandardCAN',
    kind='BO_',
    type_name='ENUM',
    choices=['StandardCAN', 'ExtendedCAN',
             'reserved', 'J1939PG',
             'reserved', 'reserved',
             'reserved', 'reserved',
             'reserved', 'reserved',
             'reserved', 'reserved',
             'reserved', 'reserved',
             'StandardCAN_FD', 'ExtendedCAN_FD'])

ATTRIBUTE_DEFINITION_CANFD_BRS = AttributeDefinition(
    name='CANFD_BRS',
    default_value='1',
    kind='BO_',
    type_name='ENUM',
    choices=['0', '1'])

ATTRIBUTE_DEFINITION_BUS_TYPE = AttributeDefinition(
    name='BusType',
    default_value='CAN',
    type_name='STRING')

ATTRIBUTE_DEFINITION_GENMSGCYCLETIME = AttributeDefinition(
    name='GenMsgCycleTime',
    default_value=0,
    kind='BO_',
    type_name='INT',
    minimum=0,
    maximum=2**16-1)

ATTRIBUTE_DEFINITION_GENSIGSTARTVALUE = AttributeDefinition(
    name='GenSigStartValue',
    default_value=0,
    kind='SG_',
    type_name='FLOAT',
    minimum=0,
    maximum=100000000000)


def to_int(value):
    return int(Decimal(value))

def to_float(value):
    return float(Decimal(value))

class Parser(textparser.Parser):

    def tokenize(self, string):
        keywords = {
            'BA_',
            'BA_DEF_',
            'BA_DEF_DEF_',
            'BA_DEF_DEF_REL_',
            'BA_DEF_REL_',
            'BA_DEF_SGTYPE_',
            'BA_REL_',
            'BA_SGTYPE_',
            'BO_',
            'BO_TX_BU_',
            'BS_',
            'BU_',
            'BU_BO_REL_',
            'BU_EV_REL_',
            'BU_SG_REL_',
            'CAT_',
            'CAT_DEF_',
            'CM_',
            'ENVVAR_DATA_',
            'EV_',
            'EV_DATA_',
            'FILTER',
            'NS_',
            'NS_DESC_',
            'SG_',
            'SG_MUL_VAL_',
            'SGTYPE_',
            'SGTYPE_VAL_',
            'SIG_GROUP_',
            'SIG_TYPE_REF_',
            'SIG_VALTYPE_',
            'SIGTYPE_VALTYPE_',
            'VAL_',
            'VAL_TABLE_',
            'VERSION'
        }

        names = {
            'LPAREN': '(',
            'RPAREN': ')',
            'LBRACE': '[',
            'RBRACE': ']',
            'COMMA':  ',',
            'AT':     '@',
            'SCOLON': ';',
            'COLON':  ':',
            'PIPE':   '|',
            'SIGN':   '+/-'
        }

        token_specs = [
            ('SKIP',     r'[ \r\n\t]+|//.*?\n'),
            ('NUMBER',   r'[-+]?\d+\.?\d*([eE][+-]?\d+)?'),
            ('WORD',     r'[A-Za-z0-9_]+'),
            ('STRING',   r'"(\\"|[^"])*?"'),
            ('LPAREN',   r'\('),
            ('RPAREN',   r'\)'),
            ('LBRACE',   r'\['),
            ('RBRACE',   r'\]'),
            ('COMMA',    r','),
            ('PIPE',     r'\|'),
            ('AT',       r'@'),
            ('SIGN',     r'[+-]'),
            ('SCOLON',   r';'),
            ('COLON',    r':'),
            ('MISMATCH', r'.')
        ]

        tokens, token_regex = tokenize_init(token_specs)

        for mo in re.finditer(token_regex, string, re.DOTALL):
            kind = mo.lastgroup

            if kind == 'SKIP':
                pass
            elif kind == 'STRING':
                value = mo.group(kind)[1:-1].replace('\\"', '"')
                tokens.append(Token(kind, value, mo.start()))
            elif kind != 'MISMATCH':
                value = mo.group(kind)

                if value in keywords:
                    kind = value

                if kind in names:
                    kind = names[kind]

                tokens.append(Token(kind, value, mo.start()))
            else:
                raise TokenizeError(string, mo.start())

        return tokens

    def grammar(self):
        version = Sequence('VERSION', 'STRING')

        ns = Sequence('NS_', ':', AnyUntil(Sequence(Any(), ':')))

        bs = Sequence('BS_', ':')

        nodes = Sequence('BU_', ':', ZeroOrMore('WORD'))

        signal = Sequence(
            'SG_', choice(Sequence('WORD', 'WORD'), Sequence('WORD')), ':',
            'NUMBER', '|', 'NUMBER', '@', 'NUMBER', '+/-',
            '(', 'NUMBER', ',', 'NUMBER', ')',
            '[', 'NUMBER', '|', 'NUMBER', ']',
            'STRING',
            DelimitedList('WORD'))

        message = Sequence(
            'BO_', 'NUMBER', 'WORD', ':', 'NUMBER', 'WORD', ZeroOrMore(signal))

        environment_variable = Sequence(
            'EV_', 'WORD', ':', 'NUMBER',
            '[', 'NUMBER', '|', 'NUMBER', ']',
            'STRING', 'NUMBER', 'NUMBER', 'WORD', 'WORD', ';')

        comment = Sequence(
            'CM_',
            choice(
                Sequence('SG_', 'NUMBER', 'WORD', 'STRING'),
                Sequence('BO_', 'NUMBER', 'STRING'),
                Sequence('EV_', 'WORD', 'STRING'),
                Sequence('BU_', 'WORD', 'STRING'),
                'STRING'),
            ';')

        attribute_definition = Sequence(
            'BA_DEF_',
            Optional(choice('SG_', 'BO_', 'EV_', 'BU_')),
            'STRING',
            'WORD',
            Optional(choice(DelimitedList('STRING'), ZeroOrMore('NUMBER'))),
            ';')

        attribute_definition_default = Sequence(
            'BA_DEF_DEF_', 'STRING', choice('NUMBER', 'STRING'), ';')

        attribute = Sequence(
            'BA_', 'STRING',
            ZeroOrMore(choice(Sequence('BO_', 'NUMBER'),
                              Sequence('SG_', 'NUMBER', 'WORD'),
                              Sequence('BU_', 'WORD'),
                              Sequence('EV_', 'WORD'))),
            choice('NUMBER', 'STRING'),
            ';')

        attribute_definition_rel = Sequence(
            'BA_DEF_REL_',
            Optional(choice('BU_SG_REL_', 'BU_BO_REL_')),
            'STRING',
            'WORD',
            Optional(choice(DelimitedList('STRING'), OneOrMore('NUMBER'))),
            ';')

        attribute_definition_default_rel = Sequence(
            'BA_DEF_DEF_REL_', 'STRING', choice('NUMBER', 'STRING'), ';')

        attribute_rel_sg = Sequence(
            'BA_REL_', 'STRING', 'BU_SG_REL_', 'WORD', 'SG_', 'NUMBER',
            'WORD', choice('NUMBER', 'STRING'), ';')

        attribute_rel_bo = Sequence(
            'BA_REL_', 'STRING', 'BU_BO_REL_', 'WORD', 'NUMBER',
            choice('NUMBER', 'STRING'), ';')

        choice_ = Sequence(
            'VAL_',
            Optional('NUMBER'),
            'WORD',
            ZeroOrMore(Sequence('NUMBER', 'STRING')),
            ';')

        value_table = Sequence(
            'VAL_TABLE_', 'WORD', ZeroOrMore(Sequence('NUMBER', 'STRING')), ';')

        signal_type = Sequence(
            'SIG_VALTYPE_', 'NUMBER', 'WORD', ':', 'NUMBER', ';')

        signal_multiplexer_values = Sequence(
            'SG_MUL_VAL_',
            'NUMBER',
            'WORD',
            'WORD',
            DelimitedList(Sequence('NUMBER', 'NUMBER')),
            ';')

        message_add_sender = Sequence(
            'BO_TX_BU_', 'NUMBER', ':', DelimitedList('WORD'), ';')

        signal_group = Sequence(
            'SIG_GROUP_', 'NUMBER', 'WORD', 'NUMBER', ':', ZeroOrMore('WORD'), ';')

        return OneOrMoreDict(
            choice(
                message,
                comment,
                attribute_definition,
                value_table,
                choice_,
                attribute,
                attribute_rel_sg,
                attribute_rel_bo,
                attribute_definition_rel,
                attribute_definition_default,
                attribute_definition_default_rel,
                signal_group,
                signal_type,
                signal_multiplexer_values,
                message_add_sender,
                environment_variable,
                nodes,
                ns,
                bs,
                version))


class LongNamesConverter:
    def __init__(self, long_names: list[str]) -> None:

        self.long_to_short: dict[str, str] = {}
        self.short_to_long: dict[str, str] = {}

        for long_name in sorted(long_names, key=lambda s: (len(s), s)):
            short_name = long_name[:32]
            index = -1
            while short_name in self.short_to_long:
                index += 1
                short_name = f'{short_name[:27]}_{index:04d}'

            self.long_to_short[long_name] = short_name
            self.short_to_long[short_name] = long_name


def get_dbc_frame_id(message):
    frame_id = message.frame_id

    if message.is_extended_frame:
        frame_id |= 0x80000000

    return frame_id

def get_dbc_name(name):
    #replace special chars with '_'
    name = re.sub(r'\W', '_', name)
    #append '_' if it starts with a number
    if name[0].isdigit():
        name = '_' + name

    return name


def _get_node_name(attributes, name):
    try:
        return attributes['node'][name]['SystemNodeLongSymbol'].value
    except (KeyError, TypeError):
        return name


def _get_environment_variable_name(attributes, name):
    try:
        return attributes['envvar'][name]['SystemEnvVarLongSymbol'].value
    except (KeyError, TypeError):
        return name


def _dump_version(database):
    return '' if database.version is None else database.version


def _dump_nodes(database):
    bu = []

    for node in database.nodes:
        bu.append(node.name)

    return bu


def _dump_value_tables(database):
    if database.dbc is None:
        return []

    val_table = []

    for name, choices in database.dbc.value_tables.items():
        choices = [
            f'{number} "{text}"'
            for number, text in sorted(choices.items(), reverse=True)
        ]
        val_table.append('VAL_TABLE_ {} {} ;'.format(name, ' '.join(choices)))

    return [*val_table, '']


def _dump_messages(database, sort_signals):
    bo = []

    def format_mux(signal):
        if signal.is_multiplexer:
            return ' M'
        elif signal.multiplexer_ids is not None:
            return f' m{signal.multiplexer_ids[0]}'
        else:
            return ''

    def format_receivers(signal):
        if signal.receivers:
            return ' ' + ','.join(signal.receivers)
        else:
            return 'Vector__XXX'

    def format_senders(message):
        if message.senders:
            return message.senders[0]
        else:
            return 'Vector__XXX'

    for message in database.messages:
        msg = []
        msg.append(
            f'BO_ {get_dbc_frame_id(message)} {message.name}: {message.length} {format_senders(message)}')

        if sort_signals:
            signals = sort_signals(message.signals)
        else:
            signals = message.signals
        for signal in signals:
            fmt = (' SG_ {name}{mux} : {start}|{length}@{byte_order}{sign}'
                   ' ({scale},{offset})'
                   ' [{minimum}|{maximum}] "{unit}" {receivers}')
            msg.append(fmt.format(
                name=signal.name,
                mux=format_mux(signal),
                start=signal.start,
                length=signal.length,
                receivers=format_receivers(signal),
                byte_order=(0 if signal.byte_order == 'big_endian' else 1),
                sign=('-' if signal.is_signed else '+'),
                scale=signal.scale,
                offset=signal.offset,
                minimum=(0 if signal.minimum is None else signal.minimum),
                maximum=(0 if signal.maximum is None else signal.maximum),
                unit='' if signal.unit is None else signal.unit))

        bo.append('\r\n'.join(msg))

    return bo


def _dump_senders(database):
    bo_tx_bu = []

    for message in database.messages:
        if len(message.senders) > 1:
            bo_tx_bu.append(
                'BO_TX_BU_ {frame_id} : {senders};'.format(
                    frame_id=get_dbc_frame_id(message),
                    senders=','.join(message.senders)))

    return bo_tx_bu


def _dump_comments(database, sort_signals):
    cm = []

    for bus in database.buses:
        if bus.comment is not None:
            cm.append(f'CM_ "{bus.comment}";')

    for node in database.nodes:
        if node.comment is not None:
            cm.append(
                'CM_ BU_ {name} "{comment}";'.format(
                    name=node.name,
                    comment=node.comment.replace('"', '\\"')))

    for message in database.messages:
        if message.comment is not None:
            cm.append(
                'CM_ BO_ {frame_id} "{comment}";'.format(
                    frame_id=get_dbc_frame_id(message),
                    comment=message.comment.replace('"', '\\"')))

        if sort_signals:
            signals = sort_signals(message.signals)
        else:
            signals = message.signals
        for signal in signals:
            if signal.comment is not None:
                cm.append(
                    'CM_ SG_ {frame_id} {name} "{comment}";'.format(
                        frame_id=get_dbc_frame_id(message),
                        name=signal.name,
                        comment=signal.comment.replace('"', '\\"')))

    if database.dbc is not None:
        # Dump environment variable comments (CM_ EV_ <name> "comment";)
        for env in database.dbc.environment_variables.values():
            if env.comment is not None:
                escaped_comment = env.comment.replace('"', '\\"')
                cm.append(f'CM_ EV_ {env.name} "{escaped_comment}";')

    return cm


def _dump_signal_types(database):
    valtype = []

    for message in database.messages:
        for signal in message.signals:
            if not signal.is_float:
                continue

            valtype.append(
                f'SIG_VALTYPE_ {get_dbc_frame_id(message)} {signal.name} : {FLOAT_LENGTH_TO_SIGNAL_TYPE[signal.length]};')

    return valtype


def _need_startval_def(database):
    return any(s.raw_initial is not None
               for m in database.messages
               for s in m.signals)

def _need_cycletime_def(database):
    # If the user has added cycle times to a database which didn't start with them,
    # we need to add the global attribute definition so the output DBC is valid
    return any(m.cycle_time is not None
               for m in database.messages)

def _bus_is_canfd(database: InternalDatabase) -> bool:
    if database.dbc is None or database.dbc.attributes is None:
        return False
    bus_type = database.dbc.attributes.get('BusType', None)
    if bus_type is None:
        return False
    return bus_type.value == 'CAN FD'  # type: ignore[no-any-return]

def _dump_attribute_definitions(database: InternalDatabase) -> list[str]:
    ba_def = []

    if database.dbc is None:
        definitions = OrderedDict()
    else:
        definitions = database.dbc.attribute_definitions

    # define "GenMsgCycleTime" attribute for specifying the cycle
    # times of messages if it has not been explicitly defined
    if 'GenMsgCycleTime' not in definitions and _need_cycletime_def(database):
        definitions['GenMsgCycleTime'] = ATTRIBUTE_DEFINITION_GENMSGCYCLETIME
    if 'GenSigStartValue' not in definitions and _need_startval_def(database):
        definitions['GenSigStartValue'] = ATTRIBUTE_DEFINITION_GENSIGSTARTVALUE

    # create 'VFrameFormat' and 'CANFD_BRS' attribute definitions if bus is CAN FD
    if _bus_is_canfd(database):
        if 'VFrameFormat' not in definitions:
            definitions['VFrameFormat'] = ATTRIBUTE_DEFINITION_VFRAMEFORMAT
        if 'CANFD_BRS' not in definitions:
            definitions['CANFD_BRS'] = ATTRIBUTE_DEFINITION_CANFD_BRS

    def get_value(definition, value):
        if definition.minimum is None:
            value = ''
        else:
            value = f' {value}'

        return value

    def get_minimum(definition):
        return get_value(definition, definition.minimum)

    def get_maximum(definition):
        return get_value(definition, definition.maximum)

    def get_kind(definition):
        return '' if definition.kind is None else definition.kind + ' '

    for definition in definitions.values():
        if definition.type_name == 'ENUM':
            choices = ','.join([f'"{choice}"'
                                for choice in definition.choices])
            ba_def.append(
                f'BA_DEF_ {get_kind(definition)} "{definition.name}" {definition.type_name}  {choices};')
        elif definition.type_name in ['INT', 'FLOAT', 'HEX']:
            ba_def.append(
                f'BA_DEF_ {get_kind(definition)} "{definition.name}" {definition.type_name}{get_minimum(definition)}{get_maximum(definition)};')
        elif definition.type_name == 'STRING':
            ba_def.append(
                f'BA_DEF_ {get_kind(definition)} "{definition.name}" {definition.type_name} ;')

    return ba_def


def _dump_attribute_definitions_rel(database):
    ba_def_rel = []

    if database.dbc is None:
        definitions = OrderedDict()
    else:
        definitions = database.dbc.attribute_definitions_rel

    def get_value(definition, value):
        if definition.minimum is None:
            value = ''
        else:
            value = f' {value}'

        return value

    def get_minimum(definition):
        return get_value(definition, definition.minimum)

    def get_maximum(definition):
        return get_value(definition, definition.maximum)

    for definition in definitions.values():
        if definition.type_name == 'ENUM':
            choices = ','.join([f'"{choice}"'
                                for choice in definition.choices])
            ba_def_rel.append(
                f'BA_DEF_REL_ {definition.kind}  "{definition.name}" {definition.type_name}  {choices};')
        elif definition.type_name in ['INT', 'FLOAT', 'HEX']:
            ba_def_rel.append(
                f'BA_DEF_REL_ {definition.kind}  "{definition.name}" {definition.type_name}{get_minimum(definition)}{get_maximum(definition)};')
        elif definition.type_name == 'STRING':
            ba_def_rel.append(
                f'BA_DEF_REL_ {definition.kind}  "{definition.name}" {definition.type_name} ;')

    return ba_def_rel


def _dump_attribute_definition_defaults(database):
    ba_def_def = []

    if database.dbc is None:
        definitions = OrderedDict()
    else:
        definitions = database.dbc.attribute_definitions

    for definition in definitions.values():
        if definition.default_value is not None:
            if definition.type_name in ["STRING", "ENUM"]:
                fmt = 'BA_DEF_DEF_  "{name}" "{value}";'
            else:
                fmt = 'BA_DEF_DEF_  "{name}" {value};'

            ba_def_def.append(fmt.format(name=definition.name,
                                         value=definition.default_value))

    return ba_def_def


def _dump_attribute_definition_defaults_rel(database):
    ba_def_def_rel = []

    if database.dbc is None:
        definitions = OrderedDict()
    else:
        definitions = database.dbc.attribute_definitions_rel

    for definition in definitions.values():
        if definition.default_value is not None:
            if definition.type_name in ["STRING", "ENUM"]:
                fmt = 'BA_DEF_DEF_REL_ "{name}" "{value}";'
            else:
                fmt = 'BA_DEF_DEF_REL_ "{name}" {value};'

            ba_def_def_rel.append(fmt.format(name=definition.name,
                                             value=definition.default_value))

    return ba_def_def_rel


def _dump_attributes(database, sort_signals, sort_attributes):
    attributes = []

    def get_value(attribute):
        result = attribute.value

        if attribute.definition.type_name == "STRING":
            result = f'"{attribute.value}"'

        return result

    if database.dbc is not None:
        if database.dbc.attributes is not None:
            for attribute in database.dbc.attributes.values():
                attributes.append(('dbc', attribute, None, None, None, None))

        for envvar in database.dbc.environment_variables.values():
            if envvar.dbc is not None:
                if envvar.dbc.attributes is not None:
                    for attribute in envvar.dbc.attributes.values():
                        attributes.append(('envvar', attribute, None, None, None, envvar))

    for node in database.nodes:
        if node.dbc is not None:
            if node.dbc.attributes is not None:
                for attribute in node.dbc.attributes.values():
                    attributes.append(('node', attribute, node, None, None, None))

    for message in database.messages:
        # retrieve the ordered dictionary of message attributes
        msg_attributes = OrderedDict()
        if message.dbc is not None and message.dbc.attributes is not None:
            msg_attributes.update(message.dbc.attributes)

        # synchronize the attribute for the message cycle time with
        # the cycle time specified by the message object
        gen_msg_cycle_time_def: AttributeDefinition  # type: ignore[annotation-unchecked]
        msg_cycle_time = message.cycle_time or 0
        if gen_msg_cycle_time_def := database.dbc.attribute_definitions.get("GenMsgCycleTime"):
            if msg_cycle_time != gen_msg_cycle_time_def.default_value:
                msg_attributes['GenMsgCycleTime'] = Attribute(
                    value=msg_cycle_time,
                    definition=gen_msg_cycle_time_def,
                )
            elif 'GenMsgCycleTime' in msg_attributes:
                del msg_attributes['GenMsgCycleTime']
        elif 'GenMsgCycleTime' in msg_attributes:
            del msg_attributes['GenMsgCycleTime']

        # if bus is CAN FD, set VFrameFormat
        v_frame_format_def: AttributeDefinition  # type: ignore[annotation-unchecked]
        if v_frame_format_def := database.dbc.attribute_definitions.get("VFrameFormat"):
            if message.protocol == 'j1939':
                v_frame_format_str = 'J1939PG'
            elif message.is_fd and message.is_extended_frame:
                v_frame_format_str = 'ExtendedCAN_FD'
            elif message.is_fd:
                v_frame_format_str = 'StandardCAN_FD'
            elif message.is_extended_frame:
                v_frame_format_str = 'ExtendedCAN'
            else:
                v_frame_format_str = 'StandardCAN'
            v_frame_format_def = _get_enum_vframeformat_attribute(v_frame_format_def)
            # only set the VFrameFormat if it valid according to the attribute definition
            if (
                v_frame_format_str in v_frame_format_def.choices
                and v_frame_format_str != v_frame_format_def.default_value
            ):
                msg_attributes['VFrameFormat'] = Attribute(
                    value=v_frame_format_def.choices.index(v_frame_format_str),
                    definition=v_frame_format_def,
                )


        # output all message attributes
        for attribute in msg_attributes.values():
            attributes.append(('message', attribute, None, message, None, None))

        # handle the signals contained in the message
        if sort_signals:
            signals = sort_signals(message.signals)
        else:
            signals = message.signals
        for signal in signals:
            # retrieve the ordered dictionary of signal attributes
            sig_attributes = OrderedDict()
            if signal.dbc is not None and signal.dbc.attributes is not None:
                sig_attributes = signal.dbc.attributes

            # synchronize the attribute for the signal start value with
            # the start value specified by the message object
            if signal.raw_initial is None and 'GenSigStartValue' in sig_attributes:
                del sig_attributes['GenSigStartValue']
            elif signal.raw_initial is not None:
                sig_attributes['GenSigStartValue'] = Attribute(
                    value=signal.raw_initial,
                    definition=ATTRIBUTE_DEFINITION_GENSIGSTARTVALUE)

            # output all signal attributes
            for attribute in sig_attributes.values():
                attributes.append(('signal', attribute, None, message, signal, None))

    if sort_attributes:
        attributes = sort_attributes(attributes)

    ba = []
    for typ, attribute, node, message, signal, envvar in attributes:
        if typ == 'dbc':
            ba.append(f'BA_ "{attribute.definition.name}" '
                      f'{get_value(attribute)};')
        elif typ == 'envvar':
            ba.append(f'BA_ "{attribute.definition.name}" '
                      f'{attribute.definition.kind} '
                      f'{envvar.name} '
                      f'{get_value(attribute)};')
        elif typ == 'node':
            ba.append(f'BA_ "{attribute.definition.name}" '
                      f'{attribute.definition.kind} '
                      f'{node.name} '
                      f'{get_value(attribute)};')
        elif typ == 'message':
            ba.append(f'BA_ "{attribute.definition.name}" '
                      f'{attribute.definition.kind} '
                      f'{get_dbc_frame_id(message)} '
                      f'{get_value(attribute)};')
        elif typ == 'signal':
            ba.append(f'BA_ "{attribute.definition.name}" '
                      f'{attribute.definition.kind} '
                      f'{get_dbc_frame_id(message)} '
                      f'{signal.name} '
                      f'{get_value(attribute)};')

    return ba


def _dump_attributes_rel(database, sort_signals):
    ba_rel = []

    def get_value(attribute):
        result = attribute.value

        if attribute.definition.type_name == "STRING":
            result = '"' + attribute.value + '"'

        return result

    if database.dbc is not None and database.dbc.attributes_rel is not None:
        attributes_rel = database.dbc.attributes_rel
        for frame_id, element in attributes_rel.items():
            if "signal" in element:
                for signal_name, signal_lst in element['signal'].items():
                    for node_name, node_dict in signal_lst['node'].items():
                        for attribute in node_dict.values():
                            ba_rel.append(f'BA_REL_ "{attribute.definition.name}" '
                                          f'BU_SG_REL_ '
                                          f'{node_name} '
                                          f'SG_ '
                                          f'{frame_id} '
                                          f'{signal_name} '
                                          f'{get_value(attribute)};')
            elif "node" in element:
                for node_name, node_dict in element['node'].items():
                    for attribute in node_dict.values():
                        ba_rel.append(f'BA_REL_ "{attribute.definition.name}" '
                                      f'BU_BO_REL_ '
                                      f'{node_name} '
                                      f'{frame_id} '
                                      f'{get_value(attribute)};')

    return ba_rel


def _dump_choices(database, sort_signals, sort_choices):
    val = []

    for message in database.messages:
        if sort_signals:
            signals = sort_signals(message.signals)
        else:
            signals = message.signals
        for signal in signals:
            if signal.choices is None:
                continue

            if sort_choices:
                choices = sort_choices(signal.choices)
            else:
                choices = signal.choices

            val.append(
                'VAL_ {frame_id} {name} {choices} ;'.format(
                    frame_id=get_dbc_frame_id(message),
                    name=signal.name,
                    choices=' '.join([f'{value} "{text}"' for value, text in choices.items()])))

    return val


def _dump_signal_groups(database):
    sig_group = []

    for message in database.messages:
        if message.signal_groups is None:
            continue

        for signal_group in message.signal_groups:
            all_sig_names = [sig.name for sig in message.signals]
            signal_group.signal_names = list(filter(lambda sig_name: sig_name in all_sig_names, signal_group.signal_names))
            sig_group.append(
                'SIG_GROUP_ {frame_id} {signal_group_name} {repetitions} : {signal_names};'.format(
                    frame_id=get_dbc_frame_id(message),
                    signal_group_name=signal_group.name,
                    repetitions=signal_group.repetitions,
                    signal_names=' '.join(signal_group.signal_names)
                ))

    return sig_group


def _is_extended_mux_needed(messages):
    """Check for messages with more than one mux signal or signals with
    more than one multiplexer value.

    """

    for message in messages:
        multiplexers = [
            signal.name
            for signal in message.signals
            if signal.is_multiplexer
        ]

        if len(multiplexers) > 1:
            return True

        for signal in message.signals:
            if signal.multiplexer_ids:
                if len(signal.multiplexer_ids) > 1:
                    return True

    return False


def _create_mux_ranges(multiplexer_ids):
    """Create a list of ranges based on a list of single values.

    Example:
        Input:  [1, 2, 3, 5,      7, 8, 9]
        Output: [[1, 3], [5, 5], [7, 9]]

    """

    ordered = sorted(multiplexer_ids)
    # Anything but ordered[0] - 1
    prev_value = ordered[0]
    ranges = []

    for value in ordered:
        if value == prev_value + 1:
            ranges[-1][1] = value
        else:
            ranges.append([value, value])

        prev_value = value

    return ranges


def _dump_signal_mux_values(database):
    """Create multiplex entries ("SG_MUL_VAL_") if extended multiplexing
    is used.

    """

    if not _is_extended_mux_needed(database.messages):
        return []

    sig_mux_values = []

    for message in database.messages:
        for signal in message.signals:
            if not signal.multiplexer_ids:
                continue

            ranges = ', '.join([
                f'{minimum}-{maximum}'
                for minimum, maximum in _create_mux_ranges(signal.multiplexer_ids)
            ])

            sig_mux_values.append(
                f'SG_MUL_VAL_ {get_dbc_frame_id(message)} {signal.name} {signal.multiplexer_signal} {ranges};')

    return sig_mux_values


def _dump_environment_variables(database: InternalDatabase) -> list[str]:
    """Dump environment variables (EV_ entries)."""
    ev_lines: list[str] = []

    if database.dbc is None:
        return ev_lines

    for env in database.dbc.environment_variables.values():
        # Prepare values, using empty strings for None where appropriate
        env_type = env.env_type if env.env_type is not None else ''
        minimum = '' if env.minimum is None else env.minimum
        maximum = '' if env.maximum is None else env.maximum
        # escape unit quotes
        unit = '' if env.unit is None else env.unit.replace('"', '\\"')
        initial = '' if env.initial_value is None else env.initial_value
        env_id = '' if env.env_id is None else env.env_id
        access_type = '' if env.access_type is None else env.access_type
        access_node = '' if env.access_node is None else env.access_node

        escaped_unit = unit
        ev_lines.append(
            f'EV_ {env.name}: {env_type} [{minimum}|{maximum}] "{escaped_unit}" {initial} {env_id} {access_type} {access_node};'
        )

    return ev_lines


def _load_comments(tokens):
    comments = defaultdict(dict)

    for comment in tokens.get('CM_', []):
        if not isinstance(comment[1], list):
            # CANdb++ behaviour: all bus comments are concatenated
            existing_comment = comments['database'].get('bus', '')
            comments['database']['bus'] = existing_comment + comment[1]
            continue

        item = comment[1]
        kind = item[0]

        if kind == 'SG_':
            frame_id = int(item[1])

            if 'signal' not in comments[frame_id]:
                comments[frame_id]['signal'] = {}

            comments[frame_id]['signal'][item[2]] = item[3]
        elif kind == 'BO_':
            frame_id = int(item[1])
            comments[frame_id]['message'] = item[2]
        elif kind == 'BU_':
            node_name = item[1]
            comments[node_name] = item[2]
        elif kind == 'EV_':
            environment_variable_name = item[1]
            comments[environment_variable_name] = item[2]

    return comments


def _load_attribute_definitions(tokens):
    return tokens.get('BA_DEF_', [])


def _load_attribute_definition_defaults(tokens):
    defaults = OrderedDict()

    for default_attr in tokens.get('BA_DEF_DEF_', []):
        defaults[default_attr[1]] = default_attr[2]

    return defaults


def _load_attribute_definitions_relation(tokens):
    return tokens.get('BA_DEF_REL_', [])


def _load_attribute_definition_relation_defaults(tokens):
    defaults = OrderedDict()

    for default_attr in tokens.get('BA_DEF_DEF_REL_', []):
        defaults[default_attr[1]] = default_attr[2]

    return defaults


def _load_attributes(tokens, definitions):
    attributes = OrderedDict()
    attributes['node'] = OrderedDict()
    attributes['envvar'] = OrderedDict()

    def to_object(attribute):
        value = attribute[3]

        definition = definitions[attribute[1]]

        if definition.type_name in ['INT', 'HEX', 'ENUM']:
            value = to_int(value)
        elif definition.type_name == 'FLOAT':
            value = to_float(value)

        return Attribute(value=value,
                         definition=definition)

    for attribute in tokens.get('BA_', []):
        name = attribute[1]

        if len(attribute[2]) > 0:
            item = attribute[2][0]
            kind = item[0]

            if kind == 'SG_':
                frame_id_dbc = int(item[1])
                signal = item[2]

                if frame_id_dbc not in attributes:
                    attributes[frame_id_dbc] = {}
                    attributes[frame_id_dbc]['message'] = OrderedDict()

                if 'signal' not in attributes[frame_id_dbc]:
                    attributes[frame_id_dbc]['signal'] = OrderedDict()

                if signal not in attributes[frame_id_dbc]['signal']:
                    attributes[frame_id_dbc]['signal'][signal] = OrderedDict()

                attributes[frame_id_dbc]['signal'][signal][name] = to_object(attribute)
            elif kind == 'BO_':
                frame_id_dbc = int(item[1])

                if frame_id_dbc not in attributes:
                    attributes[frame_id_dbc] = {}
                    attributes[frame_id_dbc]['message'] = OrderedDict()

                attributes[frame_id_dbc]['message'][name] = to_object(attribute)
            elif kind == 'BU_':
                node = item[1]

                if node not in attributes['node']:
                    attributes['node'][node] = OrderedDict()

                attributes['node'][node][name] = to_object(attribute)
            elif kind == 'EV_':
                envvar = item[1]

                if envvar not in attributes['envvar']:
                    attributes['envvar'][envvar] = OrderedDict()

                attributes['envvar'][envvar][name] = to_object(attribute)
        else:
            if 'database' not in attributes:
                attributes['database'] = OrderedDict()

            attributes['database'][name] = to_object(attribute)

    return attributes


def _load_attributes_rel(tokens, definitions):
    attributes_rel = OrderedDict()

    def to_object(attribute, value):

        definition = definitions[attribute[1]]

        if definition.type_name in ['INT', 'HEX', 'ENUM']:
            value = to_int(value)
        elif definition.type_name == 'FLOAT':
            value = to_float(value)

        return Attribute(value=value,
                         definition=definition)

    for attribute in tokens.get('BA_REL_', []):
        name = attribute[1]
        rel_type = attribute[2]
        node = attribute[3]

        if rel_type == "BU_SG_REL_":

            frame_id_dbc = int(attribute[5])
            signal = attribute[6]

            if frame_id_dbc not in attributes_rel:
                attributes_rel[frame_id_dbc] = {}

            if 'signal' not in attributes_rel[frame_id_dbc]:
                attributes_rel[frame_id_dbc]['signal'] = OrderedDict()

            if signal not in attributes_rel[frame_id_dbc]['signal']:
                attributes_rel[frame_id_dbc]['signal'][signal] = OrderedDict()

            if 'node' not in attributes_rel[frame_id_dbc]['signal'][signal]:
                attributes_rel[frame_id_dbc]['signal'][signal]['node'] = OrderedDict()

            if node not in attributes_rel[frame_id_dbc]['signal'][signal]['node']:
                attributes_rel[frame_id_dbc]['signal'][signal]['node'][node] = OrderedDict()

            attributes_rel[frame_id_dbc]['signal'][signal]['node'][node][name] = to_object(attribute, attribute[7])

        elif rel_type == "BU_BO_REL_":
            frame_id_dbc = int(attribute[4])

            if frame_id_dbc not in attributes_rel:
                attributes_rel[frame_id_dbc] = {}

            if 'node' not in attributes_rel[frame_id_dbc]:
                attributes_rel[frame_id_dbc]['node'] = OrderedDict()

            if node not in attributes_rel[frame_id_dbc]['node']:
                attributes_rel[frame_id_dbc]['node'][node] = OrderedDict()

            attributes_rel[frame_id_dbc]['node'][node][name] = to_object(attribute, attribute[5])

        else:
            pass

    return attributes_rel


def _load_value_tables(tokens):
    """Load value tables, that is, choice definitions.

    """

    value_tables = OrderedDict()

    for value_table in tokens.get('VAL_TABLE_', []):
        name = value_table[1]
        choices = {int(number): NamedSignalValue(int(number), text) for number, text in value_table[2]}
        #choices = {int(number): text for number, text in value_table[2]}
        value_tables[name] = choices

    return value_tables


def _load_environment_variables(tokens, comments, attributes, attribute_definitions):
    environment_variables = OrderedDict()

    for env_var in tokens.get('EV_', []):
        short_name = env_var[1]
        long_name = _get_environment_variable_name(attributes, short_name)
        environment_variables[long_name] = EnvironmentVariable(
            name=long_name,
            env_type=int(env_var[3]),
            minimum=num(env_var[5]),
            maximum=num(env_var[7]),
            unit=env_var[9],
            initial_value=num(env_var[10]),
            env_id=int(env_var[11]),
            access_type=env_var[12],
            access_node=env_var[13],
            comment=comments.get(env_var[1], None),
            dbc_specifics=DbcSpecifics(attributes['envvar'].get(short_name, None),
                                       attribute_definitions))

    return environment_variables

def _load_choices(tokens):
    choices = defaultdict(dict)

    for _choice in tokens.get('VAL_', []):
        if len(_choice[1]) == 0:
            continue

        od = OrderedDict((int(v[0]), NamedSignalValue(int(v[0]), v[1])) for v in _choice[3])

        if len(od) == 0:
            continue

        frame_id = int(_choice[1][0])
        choices[frame_id][_choice[2]] = od

    return choices

def _load_message_senders(tokens, attributes):
    """Load additional message senders.

    """

    message_senders = defaultdict(list)

    for senders in tokens.get('BO_TX_BU_', []):
        frame_id = int(senders[1])
        message_senders[frame_id] += [
            _get_node_name(attributes, sender) for sender in senders[3]
        ]

    return message_senders


def _load_signal_types(tokens):
    """Load signal types.

    """

    signal_types = defaultdict(dict)

    for signal_type in tokens.get('SIG_VALTYPE_', []):
        frame_id = int(signal_type[1])
        signal_name = signal_type[2]
        signal_types[frame_id][signal_name] = int(signal_type[4])

    return signal_types


def _load_signal_multiplexer_values(tokens):
    """Load additional signal multiplexer values.

    """

    signal_multiplexer_values = defaultdict(dict)

    for signal_multiplexer_value in tokens.get('SG_MUL_VAL_', []):
        frame_id = int(signal_multiplexer_value[1])
        signal_name = signal_multiplexer_value[2]
        multiplexer_signal = signal_multiplexer_value[3]
        multiplexer_ids = []

        for lower, upper in signal_multiplexer_value[4]:
            lower = int(lower)
            upper = int(upper[1:])
            # ToDo: Probably store ranges as tuples to not run out of
            #       memory on huge ranges.
            multiplexer_ids.extend(range(lower, upper + 1))

        if multiplexer_signal not in signal_multiplexer_values[frame_id]:
            signal_multiplexer_values[frame_id][multiplexer_signal] = {}

        multiplexer_signal = signal_multiplexer_values[frame_id][multiplexer_signal]
        multiplexer_signal[signal_name] = multiplexer_ids

    return signal_multiplexer_values


def _load_signal_groups(tokens, attributes):
    """Load signal groups.

    """

    signal_groups = defaultdict(list)


    def get_attributes(frame_id_dbc, signal):
        """Get attributes for given signal.

        """

        try:
            return attributes[frame_id_dbc]['signal'][signal]
        except KeyError:
            return None

    def get_signal_name(frame_id_dbc, name):
        signal_attributes = get_attributes(frame_id_dbc, name)

        try:
            return signal_attributes['SystemSignalLongSymbol'].value
        except (KeyError, TypeError):
            return name

    for signal_group in tokens.get('SIG_GROUP_',[]):
        frame_id = int(signal_group[1])
        signal_names = [get_signal_name(frame_id, signal_name) for signal_name in signal_group[5]]
        signal_groups[frame_id].append(SignalGroup(name=signal_group[2],
                                                   repetitions=int(signal_group[3]),
                                                   signal_names=signal_names))

    return signal_groups


def _load_signals(tokens,
                  comments,
                  attributes,
                  definitions,
                  choices,
                  signal_types,
                  signal_multiplexer_values,
                  frame_id_dbc,
                  multiplexer_signal):
    signal_to_multiplexer = {}

    try:
        signal_multiplexer_values = signal_multiplexer_values[frame_id_dbc]

        for multiplexer_name, items in signal_multiplexer_values.items():
            for name in items:
                signal_to_multiplexer[name] = multiplexer_name
    except KeyError:
        pass

    def get_attributes(frame_id_dbc, signal):
        """Get attributes for given signal.

        """

        try:
            return attributes[frame_id_dbc]['signal'][signal]
        except KeyError:
            return None

    def get_comment(frame_id_dbc, signal):
        """Get comment for given signal.

        """

        try:
            return comments[frame_id_dbc]['signal'][signal]
        except KeyError:
            return None

    def get_choices(frame_id_dbc, signal):
        """Get choices for given signal.

        """

        try:
            return choices[frame_id_dbc][signal]
        except KeyError:
            return None

    def get_is_multiplexer(signal):
        if len(signal[1]) == 2:
            return signal[1][1].endswith('M')
        else:
            return False

    def get_multiplexer_ids(signal, multiplexer_signal):
        ids = []

        if multiplexer_signal is not None:
            if len(signal) == 2 and not signal[1].endswith('M'):
                value = signal[1][1:].rstrip('M')
                ids.append(int(value))
        else:
            multiplexer_signal = get_multiplexer_signal(signal,
                                                        multiplexer_signal)

        try:
            ids.extend(
                signal_multiplexer_values[multiplexer_signal][signal[0]])
        except KeyError:
            pass

        if ids:
            return list(set(ids))

    def get_multiplexer_signal(signal, multiplexer_signal):
        if len(signal) != 2:
            return

        if multiplexer_signal is None:
            try:
                return signal_to_multiplexer[signal[0]]
            except KeyError:
                pass
        elif signal[0] != multiplexer_signal:
            return multiplexer_signal

    def get_receivers(receivers):
        if receivers == ['Vector__XXX']:
            receivers = []

        return [_get_node_name(attributes, receiver) for receiver in receivers]

    def get_minimum(minimum, maximum):
        if minimum == maximum == '0':
            return None
        else:
            return num(minimum)

    def get_maximum(minimum, maximum):
        if minimum == maximum == '0':
            return None
        else:
            return num(maximum)

    def get_is_float(frame_id_dbc, signal):
        """Get is_float for given signal.

        """

        try:
            return signal_types[frame_id_dbc][signal] in FLOAT_SIGNAL_TYPES
        except KeyError:
            return False

    def get_signal_name(frame_id_dbc, name):
        signal_attributes = get_attributes(frame_id_dbc, name)

        try:
            return signal_attributes['SystemSignalLongSymbol'].value
        except (KeyError, TypeError):
            return name

    def get_signal_initial_value(frame_id_dbc, name):
        signal_attributes = get_attributes(frame_id_dbc, name)

        try:
            return signal_attributes['GenSigStartValue'].value
        except (KeyError, TypeError):
            return None

    def get_signal_spn(frame_id_dbc, name):
        signal_attributes = get_attributes(frame_id_dbc, name)
        if signal_attributes is not None and 'SPN' in signal_attributes:
            if (value := signal_attributes['SPN'].value) is not None:
                return value

        if definitions is not None and 'SPN' in definitions:
            return definitions['SPN'].default_value

        return None

    signals = []

    for signal in tokens:
        signals.append(
            Signal(name=get_signal_name(frame_id_dbc, signal[1][0]),
                   start=int(signal[3]),
                   length=int(signal[5]),
                   receivers=get_receivers(signal[20]),
                   byte_order=('big_endian'
                               if signal[7] == '0'
                               else 'little_endian'),
                   is_signed=(signal[8] == '-'),
                   raw_initial=get_signal_initial_value(frame_id_dbc, signal[1][0]),
                   conversion=BaseConversion.factory(
                       scale=num(signal[10]),
                       offset=num(signal[12]),
                       is_float=get_is_float(frame_id_dbc, signal[1][0]),
                       choices=get_choices(frame_id_dbc, signal[1][0]),
                   ),
                   minimum=get_minimum(signal[15], signal[17]),
                   maximum=get_maximum(signal[15], signal[17]),
                   unit=(None if signal[19] == '' else signal[19]),
                   spn=get_signal_spn(frame_id_dbc, signal[1][0]),
                   dbc_specifics=DbcSpecifics(get_attributes(frame_id_dbc, signal[1][0]),
                                              definitions),
                   comment=get_comment(frame_id_dbc,
                                       signal[1][0]),
                   is_multiplexer=get_is_multiplexer(signal),
                   multiplexer_ids=get_multiplexer_ids(signal[1],
                                                       multiplexer_signal),
                   multiplexer_signal=get_multiplexer_signal(signal[1],
                                                             multiplexer_signal)))

    return signals


def _get_enum_vframeformat_attribute(int_attribute):
    """Get VFrameFormat attribute definition as ENUM.

    """

    if int_attribute.type_name != 'INT':
        return int_attribute

    default_value = int_attribute.default_value
    enum_attribute = deepcopy(ATTRIBUTE_DEFINITION_VFRAMEFORMAT)
    enum_attribute.default_value = enum_attribute.choices[default_value]

    return enum_attribute

def _load_messages(tokens,
                   comments,
                   attributes,
                   definitions,
                   choices,
                   message_senders,
                   signal_types,
                   signal_multiplexer_values,
                   strict,
                   bus_name,
                   signal_groups,
                   sort_signals):
    """Load messages.

    """

    def get_attributes(frame_id_dbc):
        """Get attributes for given message.

        """

        try:
            return attributes[frame_id_dbc]['message']
        except KeyError:
            return None

    def get_comment(frame_id_dbc):
        """Get comment for given message.

        """

        try:
            return comments[frame_id_dbc]['message']
        except KeyError:
            return None

    def get_send_type(frame_id_dbc):
        """Get send type for a given message.

        """

        result = None
        message_attributes = get_attributes(frame_id_dbc)

        try:
            result = message_attributes['GenMsgSendType'].value

            # if definitions is enum (otherwise above value is maintained) -> Prevents ValueError
            if definitions['GenMsgSendType'].choices is not None:
                # Resolve ENUM index to ENUM text
                result = definitions['GenMsgSendType'].choices[int(result)]
        except (KeyError, TypeError):
            try:
                result = definitions['GenMsgSendType'].default_value
            except (KeyError, TypeError):
                result = None

        return result

    def get_cycle_time(frame_id_dbc):
        """Get cycle time for a given message.

        """
        message_attributes = get_attributes(frame_id_dbc)

        gen_msg_cycle_time_def = definitions.get('GenMsgCycleTime')
        if gen_msg_cycle_time_def is None:
            return None

        if message_attributes:
            gen_msg_cycle_time_attr = message_attributes.get('GenMsgCycleTime')
            if gen_msg_cycle_time_attr:
                return gen_msg_cycle_time_attr.value or None

        return gen_msg_cycle_time_def.default_value or None


    def get_frame_format(frame_id_dbc):
        """Get frame format for a given message"""

        message_attributes = get_attributes(frame_id_dbc)
        ref_definitions = definitions.get('VFrameFormat', None)
        if ref_definitions is None:
            return None

        ref_definitions = _get_enum_vframeformat_attribute(ref_definitions)

        try:
            frame_format = message_attributes['VFrameFormat'].value
            frame_format = ref_definitions.choices[frame_format]
        except (KeyError, TypeError):
            frame_format = ref_definitions.default_value

        return frame_format

    def get_protocol(frame_id_dbc):
        """Get protocol for a given message.

        """

        frame_format = get_frame_format(frame_id_dbc)

        if frame_format == 'J1939PG':
            return 'j1939'
        else:
            return None

    def get_message_name(frame_id_dbc, name):
        message_attributes = get_attributes(frame_id_dbc)

        try:
            return message_attributes['SystemMessageLongSymbol'].value
        except (KeyError, TypeError):
            return name

    def get_signal_groups(frame_id_dbc):
        try:
            return signal_groups[frame_id_dbc]
        except KeyError:
            return None

    messages = []

    for message in tokens.get('BO_', []):
        # Any message named VECTOR__INDEPENDENT_SIG_MSG contains
        # signals not assigned to any message. Cantools does not yet
        # support unassigned signals. Discard them for now.
        if message[2] == 'VECTOR__INDEPENDENT_SIG_MSG':
            continue

        # Frame id.
        frame_id_dbc = int(message[1])
        frame_id = frame_id_dbc & 0x7fffffff
        is_extended_frame = bool(frame_id_dbc & 0x80000000)
        frame_format = get_frame_format(frame_id_dbc)
        if frame_format is not None:
            is_fd = frame_format.endswith("CAN_FD")
        else:
            is_fd = False

        # Senders.
        senders = [_get_node_name(attributes, message[5])]

        for node in message_senders.get(frame_id_dbc, []):
            if node not in senders:
                senders.append(node)

        if senders == ['Vector__XXX']:
            senders = []

        # Signal multiplexing.
        multiplexer_signal = None

        for signal in message[6]:
            if len(signal[1]) == 2:
                if signal[1][1].endswith('M'):
                    if multiplexer_signal is None:
                        multiplexer_signal = signal[1][0]
                    else:
                        multiplexer_signal = None
                        break

        signals = _load_signals(message[6],
                                comments,
                                attributes,
                                definitions,
                                choices,
                                signal_types,
                                signal_multiplexer_values,
                                frame_id_dbc,
                                multiplexer_signal)

        messages.append(
            Message(frame_id=frame_id,
                    is_extended_frame=is_extended_frame,
                    name=get_message_name(frame_id_dbc, message[2]),
                    length=int(message[4], 0),
                    senders=senders,
                    send_type=get_send_type(frame_id_dbc),
                    cycle_time=get_cycle_time(frame_id_dbc),
                    dbc_specifics=DbcSpecifics(get_attributes(frame_id_dbc),
                                               definitions),
                    signals=signals,
                    comment=get_comment(frame_id_dbc),
                    strict=strict,
                    unused_bit_pattern=0xff,
                    protocol=get_protocol(frame_id_dbc),
                    bus_name=bus_name,
                    signal_groups=get_signal_groups(frame_id_dbc),
                    sort_signals=sort_signals,
                    is_fd=is_fd))

    return messages


def _load_version(tokens):
    return tokens.get('VERSION', [[None, None]])[0][1]


def _load_bus(attributes, comments):
    try:
        bus_name = attributes['database']['DBName'].value
    except KeyError:
        bus_name = ''

    try:
        bus_baudrate = attributes['database']['Baudrate'].value
    except KeyError:
        bus_baudrate = None

    try:
        bus_comment = comments['database']['bus']
    except KeyError:
        bus_comment = None

    if not any([bus_name, bus_baudrate, bus_comment]):
        return None

    return Bus(bus_name, baudrate=bus_baudrate, comment=bus_comment)


def _load_nodes(tokens, comments, attributes, definitions):
    nodes = None

    for token in tokens.get('BU_', []):
        nodes = [Node(name=_get_node_name(attributes, node),
                      comment=comments.get(node, None),
                      dbc_specifics=DbcSpecifics(attributes['node'].get(node, None),
                                                 definitions))
                 for node in token[2]]

    return nodes


def get_attribute_definition(database, name, default):
    if database.dbc is None:
        database.dbc = DbcSpecifics()

    if name not in database.dbc.attribute_definitions:
        database.dbc.attribute_definitions[name] = default

    return database.dbc.attribute_definitions[name]


def get_long_envvar_name_attribute_definition(database):
    return get_attribute_definition(database,
                                    'SystemEnvVarLongSymbol',
                                    ATTRIBUTE_DEFINITION_LONG_ENVVAR_NAME)


def get_long_node_name_attribute_definition(database):
    return get_attribute_definition(database,
                                    'SystemNodeLongSymbol',
                                    ATTRIBUTE_DEFINITION_LONG_NODE_NAME)


def get_long_message_name_attribute_definition(database):
    return get_attribute_definition(database,
                                    'SystemMessageLongSymbol',
                                    ATTRIBUTE_DEFINITION_LONG_MESSAGE_NAME)


def get_long_signal_name_attribute_definition(database):
    return get_attribute_definition(database,
                                    'SystemSignalLongSymbol',
                                    ATTRIBUTE_DEFINITION_LONG_SIGNAL_NAME)


def try_remove_attribute(dbc, name):
    try:
        dbc.attributes.pop(name)
    except (KeyError, AttributeError):
        pass


def remove_special_chars(database):
    for node in database.nodes:
        new_node_name = get_dbc_name(node.name)

        for message in database.messages:
            for index, sender in enumerate(message.senders):
                if sender == node.name:
                    message.senders[index] = new_node_name

            for signal in message.signals:
                for index, receiver in enumerate(signal.receivers):
                    if receiver == node.name:
                        signal.receivers[index] = new_node_name

        node.name = new_node_name

    for message in database.messages:
        message.name = get_dbc_name(message.name)

        for signal in message.signals:
            signal.name = get_dbc_name(signal.name)

        if message.signal_groups is not None:
            for signal_group in message.signal_groups:
                signal_group.name = get_dbc_name(signal_group.name)
                signal_group.signal_names = [get_dbc_name(sig_name) for sig_name in signal_group.signal_names]

    return database


def make_node_names_unique(database: InternalDatabase, shorten_long_names: bool) -> None:
    converter = LongNamesConverter([node.name for node in database.nodes])

    for node in database.nodes:
        long_name = node.name
        short_name = converter.long_to_short[long_name]
        try_remove_attribute(node.dbc, 'SystemNodeLongSymbol')

        if (long_name == short_name) or not shorten_long_names:
            continue

        for message in database.messages:
            for index, sender in enumerate(message.senders):
                if sender == node.name:
                    message.senders[index] = short_name

            for signal in message.signals:
                for index, receiver in enumerate(signal.receivers):
                    if receiver == node.name:
                        signal.receivers[index] = short_name

        if node.dbc is None:
            node.dbc = DbcSpecifics()

        node.dbc.attributes['SystemNodeLongSymbol'] = Attribute(
            long_name,
            get_long_node_name_attribute_definition(database))
        node.name = short_name


def make_message_names_unique(database: InternalDatabase, shorten_long_names: bool) -> None:
    converter = LongNamesConverter([message.name for message in database.messages])

    for message in database.messages:
        long_name = message.name
        short_name = converter.long_to_short[long_name]
        try_remove_attribute(message.dbc, 'SystemMessageLongSymbol')

        if (long_name == short_name) or not shorten_long_names:
            continue

        if message.dbc is None:
            message.dbc = DbcSpecifics()

        message.dbc.attributes['SystemMessageLongSymbol'] = Attribute(
            long_name,
            get_long_message_name_attribute_definition(database))
        message.name = short_name


def make_signal_names_unique(database: InternalDatabase, shorten_long_names: bool) -> None:
    for message in database.messages:
        converter = LongNamesConverter([signal.name for signal in message.signals])
        for signal in message.signals:
            long_name = signal.name
            short_name = converter.long_to_short[long_name]
            try_remove_attribute(signal.dbc, 'SystemSignalLongSymbol')

            if (long_name == short_name) or not shorten_long_names:
                continue

            if signal.dbc is None:
                signal.dbc = DbcSpecifics()

            signal.dbc.attributes['SystemSignalLongSymbol'] = Attribute(
                long_name,
                get_long_signal_name_attribute_definition(database))
            signal.name = short_name

        if shorten_long_names and message.signal_groups:
            for signal_group in message.signal_groups:
                signal_group.signal_names = [converter.long_to_short[long_name]
                                             for long_name in signal_group.signal_names]


def make_envvar_names_unique(database: InternalDatabase, shorten_long_names: bool) -> None:
    if database.dbc is None:
        return

    envvars = database.dbc.environment_variables.values()
    converter = LongNamesConverter([envvar.name for envvar in envvars])

    for envvar in envvars:
        long_name = envvar.name
        short_name = converter.long_to_short[long_name]
        try_remove_attribute(envvar.dbc, 'SystemEnvVarLongSymbol')

        if (long_name == short_name) or not shorten_long_names:
            continue

        envvar.dbc.attributes['SystemEnvVarLongSymbol'] = Attribute(
            long_name,
            get_long_envvar_name_attribute_definition(database))
        envvar.name = short_name


def make_names_unique(database: InternalDatabase, shorten_long_names: bool) -> InternalDatabase:
    """Make message, signal and node names unique and add attributes for
    their long names.

    """

    make_node_names_unique(database, shorten_long_names)
    make_message_names_unique(database, shorten_long_names)
    make_signal_names_unique(database, shorten_long_names)
    make_envvar_names_unique(database, shorten_long_names)

    return database


def dump_string(database: InternalDatabase,
                sort_signals:type_sort_signals=SORT_SIGNALS_DEFAULT,
                sort_attribute_signals:type_sort_signals=SORT_SIGNALS_DEFAULT,
                sort_attributes:type_sort_attributes=None,
                sort_choices:type_sort_choices=None,
                shorten_long_names:bool=True) -> str:
    """Format database in DBC file format.
       sort_signals defines how to sort signals in message definitions
       sort_attribute_signals defines how to sort signals in metadata -
          comments, value table definitions and attributes

    """

    if sort_signals == SORT_SIGNALS_DEFAULT:
        sort_signals = sort_signals_by_start_bit_reversed
    if sort_attribute_signals == SORT_SIGNALS_DEFAULT:
        sort_attribute_signals = sort_signals_by_start_bit_reversed

    # Make a deep copy of the database as names and attributes will be
    # modified for items with long names.
    database = deepcopy(database)

    if database.dbc is None:
        database.dbc = DbcSpecifics()

    database = remove_special_chars(database)
    database = make_names_unique(database, shorten_long_names)
    bu = _dump_nodes(database)
    val_table = _dump_value_tables(database)
    bo = _dump_messages(database, sort_signals)
    bo_tx_bu = _dump_senders(database)
    cm = _dump_comments(database, sort_attribute_signals)
    signal_types = _dump_signal_types(database)
    ba_def = _dump_attribute_definitions(database)
    ba_def_rel = _dump_attribute_definitions_rel(database)
    ba_def_def = _dump_attribute_definition_defaults(database)
    ba_def_def_rel = _dump_attribute_definition_defaults_rel(database)
    ba = _dump_attributes(database, sort_attribute_signals, sort_attributes)
    ba_rel = _dump_attributes_rel(database, sort_attribute_signals)
    val = _dump_choices(database, sort_attribute_signals, sort_choices)
    sig_group = _dump_signal_groups(database)
    sig_mux_values = _dump_signal_mux_values(database)
    ev = _dump_environment_variables(database)

    return DBC_FMT.format(version=_dump_version(database),
                          bu=' '.join(bu),
                          val_table='\r\n'.join(val_table),
                          bo='\r\n\r\n'.join(bo),
                          bo_tx_bu='\r\n'.join(bo_tx_bu),
                          ev='\r\n\r\n'.join(ev),
                          cm='\r\n'.join(cm),
                          signal_types='\r\n'.join(signal_types),
                          ba_def='\r\n'.join(ba_def),
                          ba_def_rel="".join([elem+"\r\n" for elem in ba_def_rel]),
                          ba_def_def='\r\n'.join(ba_def_def),
                          ba_def_def_rel="".join([elem+"\r\n" for elem in ba_def_def_rel]),
                          ba='\r\n'.join(ba),
                          ba_rel="".join([elem+"\r\n" for elem in ba_rel]),
                          val='\r\n'.join(val),
                          sig_group='\r\n'.join(sig_group),
                          sig_mux_values='\r\n'.join(sig_mux_values))


def get_definitions_dict(definitions, defaults):
    result = OrderedDict()

    def convert_value(definition, value):
        if definition.type_name in ['INT', 'HEX']:
            value = to_int(value)
        elif definition.type_name == 'FLOAT':
            value = to_float(value)

        return value

    for item in definitions:
        if len(item[1]) > 0:
            kind = item[1][0]
        else:
            kind = None

        definition = AttributeDefinition(name=item[2],
                                         kind=kind,
                                         type_name=item[3])
        values = item[4][0]

        if len(values) > 0:
            if definition.type_name == "ENUM":
                definition.choices = values
            elif definition.type_name in ['INT', 'FLOAT', 'HEX']:
                definition.minimum = convert_value(definition, values[0])
                definition.maximum = convert_value(definition, values[1])

        try:
            value = defaults[definition.name]
            definition.default_value = convert_value(definition, value)
        except KeyError:
            definition.default_value = None

        result[definition.name] = definition

    return result


def get_definitions_rel_dict(definitions, defaults):
    result = OrderedDict()

    def convert_value(definition, value):
        if definition.type_name in ['INT', 'HEX']:
            value = to_int(value)
        elif definition.type_name == 'FLOAT':
            value = to_float(value)

        return value

    for item in definitions:
        if len(item[1]) > 0:
            kind = item[1][0]
        else:
            kind = None

        definition = AttributeDefinition(name=item[2],
                                         kind=kind,
                                         type_name=item[3])
        values = item[4]

        if len(values) > 0:
            if definition.type_name == "ENUM":
                definition.choices = values[0]
            elif definition.type_name in ['INT', 'FLOAT', 'HEX']:
                definition.minimum = convert_value(definition, values[0][0])
                definition.maximum = convert_value(definition, values[0][1])

        try:
            value = defaults[definition.name]
            definition.default_value = convert_value(definition, value)
        except KeyError:
            definition.default_value = None

        result[definition.name] = definition

    return result


def load_string(string: str, strict: bool = True,
                sort_signals: type_sort_signals = sort_signals_by_start_bit) -> InternalDatabase:
    """Parse given string.

    """

    tokens = Parser().parse(string)

    comments = _load_comments(tokens)
    definitions = _load_attribute_definitions(tokens)
    defaults = _load_attribute_definition_defaults(tokens)
    definitions_relation = _load_attribute_definitions_relation(tokens)
    defaults_relation = _load_attribute_definition_relation_defaults(tokens)
    attribute_definitions = get_definitions_dict(definitions, defaults)
    attributes = _load_attributes(tokens, attribute_definitions)
    attribute_rel_definitions = get_definitions_rel_dict(definitions_relation, defaults_relation)
    attributes_rel = _load_attributes_rel(tokens, attribute_rel_definitions)
    bus = _load_bus(attributes, comments)
    value_tables = _load_value_tables(tokens)
    choices = _load_choices(tokens)
    message_senders = _load_message_senders(tokens, attributes)
    signal_types = _load_signal_types(tokens)
    signal_multiplexer_values = _load_signal_multiplexer_values(tokens)
    signal_groups = _load_signal_groups(tokens, attributes)
    messages = _load_messages(tokens,
                              comments,
                              attributes,
                              attribute_definitions,
                              choices,
                              message_senders,
                              signal_types,
                              signal_multiplexer_values,
                              strict,
                              bus.name if bus else None,
                              signal_groups,
                              sort_signals)
    nodes = _load_nodes(tokens, comments, attributes, attribute_definitions)
    version = _load_version(tokens)
    environment_variables = _load_environment_variables(tokens, comments, attributes, attribute_definitions)
    dbc_specifics = DbcSpecifics(attributes.get('database', None),
                                 attribute_definitions,
                                 environment_variables,
                                 value_tables,
                                 attributes_rel,
                                 attribute_rel_definitions)

    return InternalDatabase(messages,
                            nodes,
                            [bus] if bus else [],
                            version,
                            dbc_specifics)
